""" Frontend to MySQL DB AccountingDB
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import datetime
import time
import threading
import random

from DIRAC.Core.Base.DB import DB
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.Core.Utilities import List, ThreadSafe, Time, DEncode
from DIRAC.Core.Utilities.Plotting.TypeLoader import TypeLoader
from DIRAC.Core.Utilities.ThreadPool import ThreadPool

gSynchro = ThreadSafe.Synchronizer()


class AccountingDB(DB):
    def __init__(self, name="Accounting/AccountingDB", readOnly=False):
        DB.__init__(self, "AccountingDB", name)
        self.maxBucketTime = 604800  # 1 w
        self.autoCompact = False
        self.__readOnly = readOnly
        self.__doingCompaction = False
        self.__oldBucketMethod = False
        self.__doingPendingLockTime = 0
        self.__deadLockRetries = 2
        self.__queuedRecordsLock = ThreadSafe.Synchronizer()
        self.__queuedRecordsToInsert = []
        self.dbCatalog = {}
        self.dbBucketsLength = {}
        self.__keysCache = {}
        maxParallelInsertions = self.getCSOption("ParallelRecordInsertions", 10)
        self.__threadPool = ThreadPool(1, maxParallelInsertions)
        self.__threadPool.daemonize()
        self.catalogTableName = _getTableName("catalog", "Types")
        self._createTables(
            {
                self.catalogTableName: {
                    "Fields": {
                        "name": "VARCHAR(64) UNIQUE NOT NULL",
                        "keyFields": "VARCHAR(255) NOT NULL",
                        "valueFields": "VARCHAR(255) NOT NULL",
                        "bucketsLength": "VARCHAR(255) NOT NULL",
                    },
                    "PrimaryKey": "name",
                }
            }
        )
        self.__loadCatalogFromDB()
        gMonitor.registerActivity("registeradded", "Register added", "Accounting", "entries", gMonitor.OP_ACUM)
        gMonitor.registerActivity("insertiontime", "Record insertion time", "Accounting", "seconds", gMonitor.OP_MEAN)
        gMonitor.registerActivity("querytime", "Records query time", "Accounting", "seconds", gMonitor.OP_MEAN)

        self.__compactTime = datetime.time(hour=2, minute=random.randint(0, 59), second=random.randint(0, 59))
        lcd = Time.dateTime()
        lcd.replace(hour=self.__compactTime.hour + 1, minute=0, second=0)
        self.__lastCompactionEpoch = Time.toEpoch(lcd)

        self.__registerTypes()

    def __loadTablesCreated(self):
        result = self._query("show tables")
        if not result["OK"]:  # pylint: disable=invalid-sequence-index
            return result
        return S_OK([f[0] for f in result["Value"]])  # pylint: disable=invalid-sequence-index

    def autoCompactDB(self):
        self.autoCompact = True
        th = threading.Thread(target=self.__periodicAutoCompactDB)
        th.setDaemon(1)
        th.start()

    def __periodicAutoCompactDB(self):
        while self.autoCompact:
            nct = Time.dateTime()
            if nct.hour >= self.__compactTime.hour:
                nct = nct + datetime.timedelta(days=1)
            nct = nct.replace(
                hour=self.__compactTime.hour, minute=self.__compactTime.minute, second=self.__compactTime.second
            )
            self.log.info("Next db compaction", "will be at %s" % nct)
            sleepTime = Time.toEpoch(nct) - Time.toEpoch()
            time.sleep(sleepTime)
            self.compactBuckets()

    def __registerTypes(self):
        """
        Register all types
        """
        retVal = gConfig.getSections("/DIRAC/Setups")
        if not retVal["OK"]:
            return S_ERROR("Can't get a list of setups: %s" % retVal["Message"])
        setupsList = retVal["Value"]
        objectsLoaded = TypeLoader().getTypes()

        # Load the files
        for pythonClassName in sorted(objectsLoaded):
            typeClass = objectsLoaded[pythonClassName]
            for setup in setupsList:
                typeName = "%s_%s" % (setup, pythonClassName)

                typeDef = typeClass().getDefinition()
                # dbTypeName = "%s_%s" % ( setup, typeName )
                definitionKeyFields, definitionAccountingFields, bucketsLength = typeDef[1:]
                # If already defined check the similarities
                if typeName in self.dbCatalog:
                    bucketsLength.sort()
                    if bucketsLength != self.dbBucketsLength[typeName]:
                        bucketsLength = self.dbBucketsLength[typeName]
                        self.log.warn("Bucket length has changed", "for type %s" % typeName)
                    keyFields = [f[0] for f in definitionKeyFields]
                    if keyFields != self.dbCatalog[typeName]["keys"]:
                        keyFields = self.dbCatalog[typeName]["keys"]
                        self.log.error("Definition fields have changed", "Type %s" % typeName)
                    valueFields = [f[0] for f in definitionAccountingFields]
                    if valueFields != self.dbCatalog[typeName]["values"]:
                        valueFields = self.dbCatalog[typeName]["values"]
                        self.log.error("Accountable fields have changed", "Type %s" % typeName)
                # Try to re register to check all the tables are there
                retVal = self.registerType(typeName, definitionKeyFields, definitionAccountingFields, bucketsLength)
                if not retVal["OK"]:
                    self.log.error("Can't register type", "%s: %s" % (typeName, retVal["Message"]))
                # If it has been properly registered, update info
                elif retVal["Value"]:
                    # Set the timespan
                    self.dbCatalog[typeName]["dataTimespan"] = typeClass().getDataTimespan()
                    self.dbCatalog[typeName]["definition"] = {
                        "keys": definitionKeyFields,
                        "values": definitionAccountingFields,
                    }
        return S_OK()

    def __loadCatalogFromDB(self):
        retVal = self._query(
            "SELECT `name`, `keyFields`, `valueFields`, `bucketsLength` FROM `%s`" % self.catalogTableName
        )
        if not retVal["OK"]:
            raise Exception(retVal["Message"])
        for typesEntry in retVal["Value"]:
            typeName = typesEntry[0]
            keyFields = List.fromChar(typesEntry[1], ",")
            valueFields = List.fromChar(typesEntry[2], ",")
            bucketsLength = DEncode.decode(typesEntry[3].encode())[0]
            self.__addToCatalog(typeName, keyFields, valueFields, bucketsLength)

    def getWaitingRecordsLifeTime(self):
        """
        Get the time records can live in the IN tables without no retry
        """
        return self.getCSOption("RecordMaxWaitingTime", 86400)

    def markAllPendingRecordsAsNotTaken(self):
        """
        Mark all records to be processed as not taken
        NOTE: ONLY EXECUTE THIS AT THE BEGINNING OF THE DATASTORE SERVICE!
        """
        self.log.always("Marking all records to be processed as not taken")
        for typeName in self.dbCatalog:
            sqlTableName = _getTableName("in", typeName)
            result = self._update("UPDATE `%s` SET taken=0" % sqlTableName)
            if not result["OK"]:
                return result
        return S_OK()

    def loadPendingRecords(self):
        """
        Load all records pending to insertion and generate threaded jobs
        """
        gSynchro.lock()
        try:
            now = time.time()
            if now - self.__doingPendingLockTime <= 3600:
                return S_OK()
            self.__doingPendingLockTime = now
        finally:
            gSynchro.unlock()
        self.log.info("[PENDING] Loading pending records for insertion")
        pending = 0
        now = Time.toEpoch()
        recordsPerSlot = self.getCSOption("RecordsPerSlot", 100)
        for typeName in self.dbCatalog:
            self.log.info("[PENDING] Checking %s" % typeName)
            pendingInQueue = self.__threadPool.pendingJobs()
            emptySlots = max(0, 3000 - pendingInQueue)
            self.log.info("[PENDING] %s in the queue, %d empty slots" % (pendingInQueue, emptySlots))
            if emptySlots < 1:
                continue
            emptySlots = min(100, emptySlots)
            sqlTableName = _getTableName("in", typeName)
            sqlFields = ["id"] + self.dbCatalog[typeName]["typeFields"]
            sqlCond = (
                "WHERE taken = 0 or TIMESTAMPDIFF( SECOND, takenSince, UTC_TIMESTAMP() ) > %s"
                % self.getWaitingRecordsLifeTime()
            )
            result = self._query(
                "SELECT %s FROM `%s` %s ORDER BY id ASC LIMIT %d"
                % (", ".join(["`%s`" % f for f in sqlFields]), sqlTableName, sqlCond, emptySlots * recordsPerSlot)
            )
            if not result["OK"]:
                self.log.error(
                    "[PENDING] Error when trying to get pending records", "for %s : %s" % (typeName, result["Message"])
                )
                return result
            self.log.info("[PENDING] Got %s pending records for type %s" % (len(result["Value"]), typeName))
            dbData = result["Value"]
            idList = [str(r[0]) for r in dbData]
            # If nothing to do, continue
            if not idList:
                continue
            result = self._update(
                "UPDATE `%s` SET taken=1, takenSince=UTC_TIMESTAMP() WHERE id in (%s)"
                % (sqlTableName, ", ".join(idList))
            )
            if not result["OK"]:
                self.log.error(
                    "[PENDING] Error when trying set state to waiting records",
                    "for %s : %s" % (typeName, result["Message"]),
                )
                self.__doingPendingLockTime = 0
                return result
            # Group them in groups of 10
            recordsToProcess = []
            for record in dbData:
                pending += 1
                iD = record[0]
                startTime = record[-2]
                endTime = record[-1]
                valuesList = list(record[1:-2])
                recordsToProcess.append((iD, typeName, startTime, endTime, valuesList, now))
                if len(recordsToProcess) % recordsPerSlot == 0:
                    self.__threadPool.generateJobAndQueueIt(self.__insertFromINTable, args=(recordsToProcess,))
                    recordsToProcess = []
            if recordsToProcess:
                self.__threadPool.generateJobAndQueueIt(self.__insertFromINTable, args=(recordsToProcess,))
        self.log.info("[PENDING] Got %s records requests for all types" % pending)
        self.__doingPendingLockTime = 0
        return S_OK()

    def __addToCatalog(self, typeName, keyFields, valueFields, bucketsLength):
        """
        Add type to catalog
        """
        self.log.verbose("Adding to catalog type %s" % typeName, "with length %s" % str(bucketsLength))
        self.dbCatalog[typeName] = {
            "keys": keyFields,
            "values": valueFields,
            "typeFields": [],
            "bucketFields": [],
            "dataTimespan": 0,
        }
        self.dbCatalog[typeName]["typeFields"].extend(keyFields)
        self.dbCatalog[typeName]["typeFields"].extend(valueFields)
        self.dbCatalog[typeName]["bucketFields"] = list(self.dbCatalog[typeName]["typeFields"])
        self.dbCatalog[typeName]["typeFields"].extend(["startTime", "endTime"])
        self.dbCatalog[typeName]["bucketFields"].extend(["entriesInBucket", "startTime", "bucketLength"])
        self.dbBucketsLength[typeName] = bucketsLength
        # ADRI: TEST COMPACT BUCKETS
        # self.dbBucketsLength[ typeName ] = [ ( 31104000, 3600 ) ]

    def changeBucketsLength(self, typeName, bucketsLength):
        gSynchro.lock()
        try:
            if typeName not in self.dbCatalog:
                return S_ERROR("%s is not a valid type name" % typeName)
            bucketsLength.sort()
            bucketsEncoding = DEncode.encode(bucketsLength)
            retVal = self._update(
                "UPDATE `%s` set bucketsLength = '%s' where name = '%s'"
                % (self.catalogTableName, bucketsEncoding, typeName)
            )
            if not retVal["OK"]:
                return retVal
            self.dbBucketsLength[typeName] = bucketsLength
        finally:
            gSynchro.unlock()
        return self.regenerateBuckets(typeName)

    @gSynchro
    def registerType(self, name, definitionKeyFields, definitionAccountingFields, bucketsLength):
        """
        Register a new type
        """
        gMonitor.registerActivity(
            "registerwaiting:%s" % name,
            "Records waiting for insertion for %s" % " ".join(name.split("_")),
            "Accounting",
            "records",
            gMonitor.OP_MEAN,
        )
        gMonitor.registerActivity(
            "registeradded:%s" % name,
            "Register added for %s" % " ".join(name.split("_")),
            "Accounting",
            "entries",
            gMonitor.OP_ACUM,
        )

        result = self.__loadTablesCreated()
        if not result["OK"]:
            return result
        tablesInThere = result["Value"]
        keyFieldsList = []
        valueFieldsList = []
        for key in definitionKeyFields:
            keyFieldsList.append(key[0])
        for value in definitionAccountingFields:
            valueFieldsList.append(value[0])
        for field in definitionKeyFields:
            if field in valueFieldsList:
                return S_ERROR("Key field %s is also in the list of value fields" % field)
        for field in definitionAccountingFields:
            if field in keyFieldsList:
                return S_ERROR("Value field %s is also in the list of key fields" % field)
        for bucket in bucketsLength:
            if not isinstance(bucket, tuple):
                return S_ERROR("Length of buckets should be a list of tuples")
            if len(bucket) != 2:
                return S_ERROR("Length of buckets should have 2d tuples")
        updateDBCatalog = True
        if name in self.dbCatalog:
            updateDBCatalog = False
        tables = {}
        for key in definitionKeyFields:
            keyTableName = _getTableName("key", name, key[0])
            if keyTableName not in tablesInThere:
                self.log.info("Table for key %s has to be created" % key[0])
                tables[keyTableName] = {
                    "Fields": {"id": "INTEGER NOT NULL AUTO_INCREMENT", "value": "%s NOT NULL" % key[1]},
                    "UniqueIndexes": {"valueindex": ["value"]},
                    "PrimaryKey": "id",
                }
        # Registering type
        fieldsDict = {}
        bucketFieldsDict = {}
        inbufferDict = {"id": "BIGINT NOT NULL AUTO_INCREMENT"}
        bucketIndexes = {"startTimeIndex": ["startTime"], "bucketLengthIndex": ["bucketLength"]}
        uniqueIndexFields = ["startTime"]
        for field in definitionKeyFields:
            bucketIndexes["%sIndex" % field[0]] = [field[0]]
            uniqueIndexFields.append(field[0])
            fieldsDict[field[0]] = "INTEGER NOT NULL"
            bucketFieldsDict[field[0]] = "INTEGER NOT NULL"
            inbufferDict[field[0]] = field[1] + " NOT NULL"
        for field in definitionAccountingFields:
            fieldsDict[field[0]] = field[1] + " NOT NULL"
            bucketFieldsDict[field[0]] = "DECIMAL(30,10) NOT NULL"
            inbufferDict[field[0]] = field[1] + " NOT NULL"
        fieldsDict["startTime"] = "INT UNSIGNED NOT NULL"
        fieldsDict["endTime"] = "INT UNSIGNED NOT NULL"
        bucketFieldsDict["entriesInBucket"] = "DECIMAL(30,10) NOT NULL"
        bucketFieldsDict["startTime"] = "INT UNSIGNED NOT NULL"
        inbufferDict["startTime"] = "INT UNSIGNED NOT NULL"
        inbufferDict["endTime"] = "INT UNSIGNED NOT NULL"
        inbufferDict["taken"] = "TINYINT(1) DEFAULT 1 NOT NULL"
        inbufferDict["takenSince"] = "DATETIME NOT NULL"
        bucketFieldsDict["bucketLength"] = "MEDIUMINT UNSIGNED NOT NULL"
        uniqueIndexFields.append("bucketLength")
        bucketTableName = _getTableName("bucket", name)
        if bucketTableName not in tablesInThere:
            tables[bucketTableName] = {
                "Fields": bucketFieldsDict,
                "UniqueIndexes": {"UniqueConstraint": uniqueIndexFields},
            }
        typeTableName = _getTableName("type", name)
        if typeTableName not in tablesInThere:
            tables[typeTableName] = {"Fields": fieldsDict}
        inTableName = _getTableName("in", name)
        if inTableName not in tablesInThere:
            tables[inTableName] = {"Fields": inbufferDict, "PrimaryKey": "id"}
        if self.__readOnly:
            if tables:
                self.log.notice("ReadOnly mode: Skipping create of tables for %s. Removing from memory catalog" % name)
                self.log.verbose("Skipping creation of tables %s" % ", ".join([tn for tn in tables]))
                try:
                    self.dbCatalog.pop(name)
                except KeyError:
                    pass
            else:
                self.log.notice("ReadOnly mode: %s is OK" % name)
            return S_OK(not updateDBCatalog)

        if tables:
            retVal = self._createTables(tables)
            if not retVal["OK"]:
                self.log.error("Can't create type", "%s: %s" % (name, retVal["Message"]))
                return S_ERROR("Can't create type %s: %s" % (name, retVal["Message"]))
        if updateDBCatalog:
            bucketsLength.sort()
            bucketsEncoding = DEncode.encode(bucketsLength)
            self.insertFields(
                self.catalogTableName,
                ["name", "keyFields", "valueFields", "bucketsLength"],
                [name, ",".join(keyFieldsList), ",".join(valueFieldsList), bucketsEncoding],
            )
            self.__addToCatalog(name, keyFieldsList, valueFieldsList, bucketsLength)
        self.log.info("Registered type %s" % name)
        return S_OK(True)

    def getRegisteredTypes(self):
        """
        Get list of registered types
        """
        retVal = self._query(
            "SELECT `name`, `keyFields`, `valueFields`, `bucketsLength` FROM `%s`" % self.catalogTableName
        )
        if not retVal["OK"]:
            return retVal
        typesList = []
        for name, keyFields, valueFields, bucketsLength in retVal["Value"]:
            keyFields = List.fromChar(keyFields)
            valueFields = List.fromChar(valueFields)
            bucketsLength = DEncode.decode(bucketsLength.encode())
            typesList.append([name, keyFields, valueFields, bucketsLength])
        return S_OK(typesList)

    def getKeyValues(self, typeName, condDict, connObj=False):
        """
        Get all values for a given key field in a type
        """
        keyValuesDict = {}

        keyTables = []
        sqlCond = []
        mainTable = "`%s`" % _getTableName("bucket", typeName)
        try:
            typeKeysList = self.dbCatalog[typeName]["keys"]
        except KeyError:
            return S_ERROR("Please select a category")

        for keyName in condDict:
            if keyName in typeKeysList:
                keyTable = "`%s`" % _getTableName("key", typeName, keyName)
                if keyTable not in keyTables:
                    keyTables.append(keyTable)
                sqlCond.append("%s.id = %s.`%s`" % (keyTable, mainTable, keyName))
                for value in condDict[keyName]:
                    sqlCond.append("%s.value = %s" % (keyTable, self._escapeString(value)["Value"]))

        for keyName in typeKeysList:
            keyTable = "`%s`" % _getTableName("key", typeName, keyName)
            allKeyTables = keyTables
            if keyTable not in allKeyTables:
                allKeyTables = list(keyTables)
                allKeyTables.append(keyTable)
            cmd = "SELECT DISTINCT %s.value FROM %s" % (keyTable, ", ".join(allKeyTables))
            if sqlCond:
                sqlValueLink = "%s.id = %s.`%s`" % (keyTable, mainTable, keyName)
                cmd += ", %s WHERE %s AND %s" % (mainTable, sqlValueLink, " AND ".join(sqlCond))
            retVal = self._query(cmd, conn=connObj)
            if not retVal["OK"]:
                return retVal
            keyValuesDict[keyName] = [r[0] for r in retVal["Value"]]

        return S_OK(keyValuesDict)

    @gSynchro
    def deleteType(self, typeName):
        """
        Deletes a type
        """
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        if typeName not in self.dbCatalog:
            return S_ERROR("Type %s does not exist" % typeName)
        self.log.info("Deleting type", typeName)
        tablesToDelete = []
        for keyField in self.dbCatalog[typeName]["keys"]:
            tablesToDelete.append("`%s`" % _getTableName("key", typeName, keyField))
        tablesToDelete.insert(0, "`%s`" % _getTableName("type", typeName))
        tablesToDelete.insert(0, "`%s`" % _getTableName("bucket", typeName))
        tablesToDelete.insert(0, "`%s`" % _getTableName("in", typeName))
        retVal = self._query("DROP TABLE %s" % ", ".join(tablesToDelete))
        if not retVal["OK"]:
            return retVal
        retVal = self._update("DELETE FROM `%s` WHERE name='%s'" % (_getTableName("catalog", "Types"), typeName))
        del self.dbCatalog[typeName]
        return S_OK()

    def __getIdForKeyValue(self, typeName, keyName, keyValue, conn=False):
        """
        Finds id number for value in a key table
        """
        retVal = self._escapeString(keyValue)
        if not retVal["OK"]:
            return retVal
        keyValue = retVal["Value"]
        retVal = self._query(
            "SELECT `id` FROM `%s` WHERE `value`=%s" % (_getTableName("key", typeName, keyName), keyValue), conn=conn
        )
        if not retVal["OK"]:
            return retVal
        if len(retVal["Value"]) > 0:
            return S_OK(retVal["Value"][0][0])
        return S_ERROR("Key id %s for value %s does not exist although it shoud" % (keyName, keyValue))

    def __addKeyValue(self, typeName, keyName, keyValue):
        """
        Adds a key value to a key table if not existant
        """
        # Cast to string just in case
        if not isinstance(keyValue, six.string_types):
            keyValue = str(keyValue)
        # No more than 64 chars for keys
        if len(keyValue) > 64:
            keyValue = keyValue[:64]

        # Look into the cache
        if typeName not in self.__keysCache:
            self.__keysCache[typeName] = {}
        typeCache = self.__keysCache[typeName]
        if keyName not in typeCache:
            typeCache[keyName] = {}
        keyCache = typeCache[keyName]
        if keyValue in keyCache:
            return S_OK(keyCache[keyValue])
        # Retrieve key
        keyTable = _getTableName("key", typeName, keyName)
        retVal = self.__getIdForKeyValue(typeName, keyName, keyValue)
        if retVal["OK"]:
            keyCache[keyValue] = retVal["Value"]
            return retVal
        # Key is not in there
        retVal = self._getConnection()
        if not retVal["OK"]:
            return retVal
        connection = retVal["Value"]
        self.log.info("Value %s for key %s didn't exist, inserting" % (keyValue, keyName))
        retVal = self.insertFields(keyTable, ["id", "value"], [0, keyValue], connection)
        if not retVal["OK"] and retVal["Message"].find("Duplicate key") == -1:
            return retVal
        result = self.__getIdForKeyValue(typeName, keyName, keyValue, connection)
        if not result["OK"]:
            return result
        keyCache[keyValue] = result["Value"]
        return result

    def calculateBucketLengthForTime(self, typeName, now, when):
        """
        Get the expected bucket time for a moment in time
        """
        for granuT in self.dbBucketsLength[typeName]:
            nowBucketed = now - now % granuT[1]
            dif = max(0, nowBucketed - when)
            if dif <= granuT[0]:
                return granuT[1]
        return self.maxBucketTime

    def calculateBuckets(self, typeName, startTime, endTime, nowEpoch=False):
        """
        Magic function for calculating buckets between two times and
        the proportional part for each bucket
        """
        if not nowEpoch:
            nowEpoch = int(Time.toEpoch(Time.dateTime()))
        bucketTimeLength = self.calculateBucketLengthForTime(typeName, nowEpoch, startTime)
        currentBucketStart = startTime - startTime % bucketTimeLength
        if startTime == endTime:
            return [(currentBucketStart, 1, bucketTimeLength)]
        buckets = []
        totalLength = endTime - startTime
        while currentBucketStart < endTime:
            start = max(currentBucketStart, startTime)
            end = min(currentBucketStart + bucketTimeLength, endTime)
            proportion = float(end - start) / totalLength
            buckets.append((currentBucketStart, proportion, bucketTimeLength))
            currentBucketStart += bucketTimeLength
            bucketTimeLength = self.calculateBucketLengthForTime(typeName, nowEpoch, currentBucketStart)
        return buckets

    def __insertInQueueTable(self, typeName, startTime, endTime, valuesList):
        sqlFields = ["id", "taken", "takenSince"] + self.dbCatalog[typeName]["typeFields"]
        sqlValues = ["0", "0", "UTC_TIMESTAMP()"] + valuesList + [startTime, endTime]
        if len(sqlFields) != len(sqlValues):
            numRcv = len(valuesList) + 2
            numExp = len(self.dbCatalog[typeName]["typeFields"])
            return S_ERROR("Fields mismatch for record %s. %s fields and %s expected" % (typeName, numRcv, numExp))
        retVal = self.insertFields(_getTableName("in", typeName), sqlFields, sqlValues)
        if not retVal["OK"]:
            return retVal
        return S_OK(retVal["lastRowId"])

    def insertRecordBundleThroughQueue(self, recordsToQueue):
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        recordsToProcess = []
        now = Time.toEpoch()
        for record in recordsToQueue:
            typeName, startTime, endTime, valuesList = record
            result = self.__insertInQueueTable(typeName, startTime, endTime, valuesList)
            if not result["OK"]:
                return result
            iD = result["Value"]
            recordsToProcess.append((iD, typeName, startTime, endTime, valuesList, now))

        return S_OK()

    def insertRecordThroughQueue(self, typeName, startTime, endTime, valuesList):
        """
        Insert a record in the intable to be really insterted afterwards
        """
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        self.log.info(
            "Adding record to queue",
            "for type %s\n [%s -> %s]" % (typeName, Time.fromEpoch(startTime), Time.fromEpoch(endTime)),
        )
        if typeName not in self.dbCatalog:
            return S_ERROR("Type %s has not been defined in the db" % typeName)
        result = self.__insertInQueueTable(typeName, startTime, endTime, valuesList)
        if not result["OK"]:
            return result

        return S_OK()

    def __insertFromINTable(self, recordTuples):
        """
        Do the real insert and delete from the in buffer table
        """
        self.log.verbose("Received bundle to process", "of %s elements" % len(recordTuples))
        for record in recordTuples:
            iD, typeName, startTime, endTime, valuesList, insertionEpoch = record
            result = self.insertRecordDirectly(typeName, startTime, endTime, valuesList)
            if not result["OK"]:
                self._update("UPDATE `%s` SET taken=0 WHERE id=%s" % (_getTableName("in", typeName), iD))
                self.log.error("Can't insert row", result["Message"])
                continue
            result = self._update("DELETE FROM `%s` WHERE id=%s" % (_getTableName("in", typeName), iD))
            if not result["OK"]:
                self.log.error("Can't delete row from the IN table", result["Message"])
            gMonitor.addMark("insertiontime", Time.toEpoch() - insertionEpoch)

    def insertRecordDirectly(self, typeName, startTime, endTime, valuesList):
        """
        Add an entry to the type contents
        """
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        gMonitor.addMark("registeradded", 1)
        gMonitor.addMark("registeradded:%s" % typeName, 1)
        self.log.info(
            "Adding record", "for type %s\n [%s -> %s]" % (typeName, Time.fromEpoch(startTime), Time.fromEpoch(endTime))
        )
        if typeName not in self.dbCatalog:
            return S_ERROR("Type %s has not been defined in the db" % typeName)
        # Discover key indexes
        for keyPos, keyName in enumerate(self.dbCatalog[typeName]["keys"]):
            keyValue = valuesList[keyPos]
            retVal = self.__addKeyValue(typeName, keyName, keyValue)
            if not retVal["OK"]:
                return retVal
            self.log.verbose("Value %s for key %s has id %s" % (keyValue, keyName, retVal["Value"]))
            valuesList[keyPos] = retVal["Value"]
        insertList = list(valuesList)
        insertList.append(startTime)
        insertList.append(endTime)
        retVal = self._getConnection()
        if not retVal["OK"]:
            return retVal
        connObj = retVal["Value"]
        try:
            retVal = self.insertFields(
                _getTableName("type", typeName), self.dbCatalog[typeName]["typeFields"], insertList, conn=connObj
            )
            if not retVal["OK"]:
                return retVal
            # HACK: One more record to split in the buckets to be able to count total entries
            valuesList.append(1)
            retVal = self.__startTransaction(connObj)
            if not retVal["OK"]:
                return retVal
            retVal = self.__splitInBuckets(typeName, startTime, endTime, valuesList, connObj=connObj)
            if not retVal["OK"]:
                self.__rollbackTransaction(connObj)
                return retVal
            return self.__commitTransaction(connObj)
        finally:
            connObj.close()

    def deleteRecord(self, typeName, startTime, endTime, valuesList):
        """
        Delete an entry
        """
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        if typeName not in self.dbCatalog:
            return S_ERROR("Type %s has not been defined in the db" % typeName)

        self.log.info(
            "Deleting record",
            "for type %s\n [%s -> %s]" % (typeName, Time.fromEpoch(startTime), Time.fromEpoch(endTime)),
        )
        sqlValues = []
        sqlValues.extend(valuesList)
        # Discover key indexes
        for keyPos, keyName in enumerate(self.dbCatalog[typeName]["keys"]):
            keyValue = sqlValues[keyPos]
            retVal = self.__addKeyValue(typeName, keyName, keyValue)
            if not retVal["OK"]:
                return retVal
            self.log.verbose("Value %s for key %s has id %s" % (keyValue, keyName, retVal["Value"]))
            sqlValues[keyPos] = retVal["Value"]
        sqlCond = []
        mainTable = _getTableName("type", typeName)
        sqlValues.extend([startTime, endTime])
        numKeyFields = len(self.dbCatalog[typeName]["keys"])
        numValueFields = len(self.dbCatalog[typeName]["values"])
        for i, value in enumerate(sqlValues):
            needToRound = False
            if i >= numKeyFields and i - numKeyFields < numValueFields:
                vIndex = i - numKeyFields
                if self.dbCatalog[typeName]["definition"]["values"][vIndex][1].find("FLOAT") > -1:
                    needToRound = True
            if needToRound:
                compVal = ["`%s`.`%s`" % (mainTable, self.dbCatalog[typeName]["typeFields"][i]), "%f" % value]
                compVal = ["CEIL( %s * 1000 )" % v for v in compVal]
                compVal = "ABS( %s ) <= 1 " % " - ".join(compVal)
            else:
                sqlCond.append("`%s`.`%s`=%s" % (mainTable, self.dbCatalog[typeName]["typeFields"][i], value))
        retVal = self._getConnection()
        if not retVal["OK"]:
            return retVal
        connObj = retVal["Value"]
        retVal = self.__startTransaction(connObj)
        if not retVal["OK"]:
            return retVal
        retVal = self._update("DELETE FROM `%s` WHERE %s" % (mainTable, " AND ".join(sqlCond)), conn=connObj)
        if not retVal["OK"]:
            return retVal
        numInsertions = retVal["Value"]
        # Deleted from type, now the buckets
        # HACK: One more record to split in the buckets to be able to count total entries
        if numInsertions == 0:
            return S_OK(0)
        sqlValues.append(1)
        retVal = self.__deleteFromBuckets(typeName, startTime, endTime, sqlValues, numInsertions, connObj=connObj)
        if not retVal["OK"]:
            self.__rollbackTransaction(connObj)
            return retVal
        retVal = self.__commitTransaction(connObj)
        if not retVal["OK"]:
            self.__rollbackTransaction(connObj)
            return retVal
        return S_OK(numInsertions)

    def __splitInBuckets(self, typeName, startTime, endTime, valuesList, connObj=False):
        """
        Bucketize a record
        """
        # Calculate amount of buckets
        buckets = self.calculateBuckets(typeName, startTime, endTime)
        if not buckets:
            return S_OK()
        # Separate key values from normal values
        numKeys = len(self.dbCatalog[typeName]["keys"])
        keyValues = valuesList[:numKeys]
        valuesList = valuesList[numKeys:]
        self.log.verbose("Splitting entry", " in %s buckets" % len(buckets))
        return self.__writeBuckets(typeName, buckets, keyValues, valuesList, connObj=connObj)

    def __deleteFromBuckets(self, typeName, startTime, endTime, valuesList, numInsertions, connObj=False):
        """
        DeBucketize a record
        """
        # Calculate amount of buckets
        buckets = self.calculateBuckets(typeName, startTime, endTime, self.__lastCompactionEpoch)
        # Separate key values from normal values
        numKeys = len(self.dbCatalog[typeName]["keys"])
        keyValues = valuesList[:numKeys]
        valuesList = valuesList[numKeys:]
        self.log.verbose("Deleting bucketed entry", "from %s buckets" % len(buckets))
        for bucketInfo in buckets:
            bucketStartTime = bucketInfo[0]
            bucketProportion = bucketInfo[1]
            bucketLength = bucketInfo[2]
            for _i in range(max(1, self.__deadLockRetries)):
                retVal = self.__extractFromBucket(
                    typeName,
                    bucketStartTime,
                    bucketLength,
                    keyValues,
                    valuesList,
                    bucketProportion * numInsertions,
                    connObj=connObj,
                )
                if not retVal["OK"]:
                    # If failed because of dead lock try restarting
                    if retVal["Message"].find("try restarting transaction"):
                        continue
                    return retVal
                # If OK, break loop
                if retVal["OK"]:
                    break
        return S_OK()

    def getBucketsDef(self, typeName):
        return self.dbBucketsLength[typeName]

    def __generateSQLConditionForKeys(self, typeName, keyValues):
        """
        Generate sql condition for buckets, values are indexes to real values
        """
        realCondList = []
        for keyPos, keyField in enumerate(self.dbCatalog[typeName]["keys"]):
            keyValue = keyValues[keyPos]
            retVal = self._escapeString(keyValue)
            if not retVal["OK"]:
                return retVal
            keyValue = retVal["Value"]
            realCondList.append("`%s`.`%s` = %s" % (_getTableName("bucket", typeName), keyField, keyValue))
        return " AND ".join(realCondList)

    def __getBucketFromDB(self, typeName, startTime, bucketLength, keyValues, connObj=False):
        """
        Get a bucket from the DB
        """
        tableName = _getTableName("bucket", typeName)
        sqlFields = []
        for valueField in self.dbCatalog[typeName]["values"]:
            sqlFields.append("`%s`.`%s`" % (tableName, valueField))
        sqlFields.append("`%s`.`entriesInBucket`" % (tableName))
        cmd = "SELECT %s FROM `%s`" % (", ".join(sqlFields), _getTableName("bucket", typeName))
        cmd += " WHERE `%s`.`startTime`='%s' AND `%s`.`bucketLength`='%s' AND " % (
            tableName,
            startTime,
            tableName,
            bucketLength,
        )
        cmd += self.__generateSQLConditionForKeys(typeName, keyValues)
        return self._query(cmd, conn=connObj)

    def __extractFromBucket(
        self, typeName, startTime, bucketLength, keyValues, bucketValues, proportion, connObj=False
    ):
        """
        Update a bucket when coming from the raw insert
        """
        tableName = _getTableName("bucket", typeName)
        cmd = "UPDATE `%s` SET " % tableName
        sqlValList = []
        for pos, valueField in enumerate(self.dbCatalog[typeName]["values"]):
            value = bucketValues[pos]
            fullFieldName = "`%s`.`%s`" % (tableName, valueField)
            sqlValList.append("%s=GREATEST(0,%s-(%s*%s))" % (fullFieldName, fullFieldName, value, proportion))
        sqlValList.append(
            "`%s`.`entriesInBucket`=GREATEST(0,`%s`.`entriesInBucket`-(%s*%s))"
            % (tableName, tableName, bucketValues[-1], proportion)
        )
        cmd += ", ".join(sqlValList)
        cmd += " WHERE `%s`.`startTime`='%s' AND `%s`.`bucketLength`='%s' AND " % (
            tableName,
            startTime,
            tableName,
            bucketLength,
        )
        cmd += self.__generateSQLConditionForKeys(typeName, keyValues)
        return self._update(cmd, conn=connObj)

    def __writeBuckets(self, typeName, buckets, keyValues, valuesList, connObj=False):
        """Insert or update a bucket"""
        #     tableName = _getTableName( "bucket", typeName )
        # INSERT PART OF THE QUERY
        sqlFields = ["`startTime`", "`bucketLength`", "`entriesInBucket`"]
        for keyPos in range(len(self.dbCatalog[typeName]["keys"])):
            sqlFields.append("`%s`" % self.dbCatalog[typeName]["keys"][keyPos])
        sqlUpData = ["`entriesInBucket`=`entriesInBucket`+VALUES(`entriesInBucket`)"]
        for valPos in range(len(self.dbCatalog[typeName]["values"])):
            valueField = "`%s`" % self.dbCatalog[typeName]["values"][valPos]
            sqlFields.append(valueField)
            sqlUpData.append("%s=%s+VALUES(%s)" % (valueField, valueField, valueField))
        valuesGroups = []
        for bucketInfo in buckets:
            bStartTime = bucketInfo[0]
            bProportion = bucketInfo[1]
            bLength = bucketInfo[2]
            sqlValues = [bStartTime, bLength, "(%s*%s)" % (valuesList[-1], bProportion)]
            for keyPos in range(len(self.dbCatalog[typeName]["keys"])):
                sqlValues.append(keyValues[keyPos])
            for valPos in range(len(self.dbCatalog[typeName]["values"])):
                #         value = valuesList[ valPos ]
                sqlValues.append("(%s*%s)" % (valuesList[valPos], bProportion))
            valuesGroups.append("( %s )" % ",".join(str(val) for val in sqlValues))

        cmd = "INSERT INTO `%s` ( %s ) " % (_getTableName("bucket", typeName), ", ".join(sqlFields))
        cmd += "VALUES %s " % ", ".join(valuesGroups)
        cmd += "ON DUPLICATE KEY UPDATE %s" % ", ".join(sqlUpData)

        for _i in range(max(1, self.__deadLockRetries)):
            result = self._update(cmd, conn=connObj)
            if not result["OK"]:
                # If failed because of dead lock try restarting
                if result["Message"].find("try restarting transaction"):
                    continue
                return result
            # If OK, break loopo
            if result["OK"]:
                return result

        return S_ERROR("Cannot update bucket: %s" % result["Message"])

    def __checkFieldsExistsInType(self, typeName, fields, tableType):
        """
        Check wether a list of fields exist for a given typeName
        """
        missing = []
        tableFields = self.dbCatalog[typeName]["%sFields" % tableType]
        for key in fields:
            if key not in tableFields:
                missing.append(key)
        return missing

    def __checkIncomingFieldsForQuery(self, typeName, selectFields, condDict, groupFields, orderFields, tableType):
        missing = self.__checkFieldsExistsInType(typeName, selectFields[1], tableType)
        if missing:
            return S_ERROR("Value keys %s are not defined" % ", ".join(missing))
        missing = self.__checkFieldsExistsInType(typeName, condDict, tableType)
        if missing:
            return S_ERROR("Condition keys %s are not defined" % ", ".join(missing))
        if groupFields:
            missing = self.__checkFieldsExistsInType(typeName, groupFields[1], tableType)
            if missing:
                return S_ERROR("Group fields %s are not defined" % ", ".join(missing))
        if orderFields:
            missing = self.__checkFieldsExistsInType(typeName, orderFields[1], tableType)
            if missing:
                return S_ERROR("Order fields %s are not defined" % ", ".join(missing))
        return S_OK()

    def retrieveRawRecords(self, typeName, startTime, endTime, condDict, orderFields, connObj=False):
        """
        Get RAW data from the DB
        """
        if typeName not in self.dbCatalog:
            return S_ERROR("Type %s not defined" % typeName)
        selectFields = [["%s", "%s"], ["startTime", "endTime"]]
        for tK in ("keys", "values"):
            for key in self.dbCatalog[typeName][tK]:
                selectFields[0].append("%s")
                selectFields[1].append(key)
        selectFields[0] = ", ".join(selectFields[0])
        return self.__queryType(typeName, startTime, endTime, selectFields, condDict, False, orderFields, "type")

    def retrieveBucketedData(
        self, typeName, startTime, endTime, selectFields, condDict, groupFields, orderFields, connObj=False
    ):
        """
        Get data from the DB

        Parameters:
         - typeName -> typeName
         - startTime & endTime -> int
             epoch objects. Do I need to explain the meaning?
         - selectFields: list containing a string and a list of fields:
                           ["SUM(%s), %s/%s", ["field1name", "field2name", "field3name"]]
         - condDict -> conditions for the query
                       key -> name of the field
                       value -> list of possible values
         - groupFields -> list of fields to group by
                          ( "%s, %s, %s", ( "field1name", "field2name", "field3name" ) )
         - orderFields -> list of fields to order by
                          ( "%s, %s, %s", ( "field1name", "field2name", "field3name" ) )

        """
        if typeName not in self.dbCatalog:
            return S_ERROR("Type %s is not defined" % typeName)
        startQueryEpoch = time.time()
        if len(selectFields) < 2:
            return S_ERROR("selectFields has to be a list containing a string and a list of fields")
        retVal = self.__checkIncomingFieldsForQuery(
            typeName, selectFields, condDict, groupFields, orderFields, "bucket"
        )
        if not retVal["OK"]:
            return retVal
        nowEpoch = Time.toEpoch(Time.dateTime())
        bucketTimeLength = self.calculateBucketLengthForTime(typeName, nowEpoch, startTime)
        startTime = startTime - startTime % bucketTimeLength
        result = self.__queryType(
            typeName, startTime, endTime, selectFields, condDict, groupFields, orderFields, "bucket", connObj=connObj
        )
        gMonitor.addMark("querytime", Time.toEpoch() - startQueryEpoch)
        return result

    def __queryType(
        self, typeName, startTime, endTime, selectFields, condDict, groupFields, orderFields, tableType, connObj=False
    ):
        """
        Execute a query over a main table
        """

        tableName = _getTableName(tableType, typeName)
        cmd = "SELECT"
        sqlLinkList = []
        # Check if groupFields and orderFields are in ( "%s", ( field1, ) ) form
        if groupFields:
            try:
                groupFields[0] % tuple(groupFields[1])
                # We can have the case when we have multiple grouping and the fields in the select
                # does not much the group by conditions
                # for example: selectFields = ('%s, %s, %s, SUM(%s)', ['Site', 'startTime', 'bucketLength', 'entriesInBucket'])
                #             groupFields = ('%s, %s', ['startTime', 'Site'])
                #             in this case the correct query must be: select Site, startTime, bucketlength,
                #                               sum(entriesInBucket) from xxxx where yyy Group by Site, startTime, bucketlength
                #
                # When we have multiple grouping then we must have all the fields in Group by. This is from mysql 5.7.
                # We have fields which are not in the groupFields and it is in selectFields

                if "bucketLength" in selectFields[1]:
                    groupFields = list(groupFields)
                    groupFields[0] = "%s, %s" % (groupFields[0], "%s")
                    groupFields[1].append("bucketlength")
                    groupFields = tuple(groupFields)

            except TypeError as e:
                return S_ERROR("Cannot format properly group string: %s" % repr(e))
        if orderFields:
            try:
                orderFields[0] % tuple(orderFields[1])
            except TypeError as e:
                return S_ERROR("Cannot format properly order string: %s" % repr(e))
        # Calculate fields to retrieve
        realFieldList = []
        for rawFieldName in selectFields[1]:
            keyTable = _getTableName("key", typeName, rawFieldName)
            if rawFieldName in self.dbCatalog[typeName]["keys"]:
                realFieldList.append("`%s`.`value`" % keyTable)
                List.appendUnique(sqlLinkList, "`%s`.`%s` = `%s`.`id`" % (tableName, rawFieldName, keyTable))
            else:
                realFieldList.append("`%s`.`%s`" % (tableName, rawFieldName))
        try:
            cmd += " %s" % selectFields[0] % tuple(realFieldList)
        except TypeError as e:
            return S_ERROR("Error generating select fields string: %s" % repr(e))
        # Calculate tables needed
        sqlFromList = ["`%s`" % tableName]
        for key in self.dbCatalog[typeName]["keys"]:
            if (
                key in condDict
                or key in selectFields[1]
                or (groupFields and key in groupFields[1])
                or (orderFields and key in orderFields[1])
            ):
                sqlFromList.append("`%s`" % _getTableName("key", typeName, key))
        cmd += " FROM %s" % ", ".join(sqlFromList)
        # Calculate time conditions
        sqlTimeCond = []
        if startTime:
            if tableType == "bucket":
                # HACK because MySQL and UNIX do not start epoch at the same time
                startTime = startTime + 3600
                startTime = self.calculateBuckets(typeName, startTime, startTime)[0][0]
            sqlTimeCond.append("`%s`.`startTime` >= %s" % (tableName, startTime))
        if endTime:
            if tableType == "bucket":
                endTimeSQLVar = "startTime"
                endTime = endTime + 3600
                endTime = self.calculateBuckets(typeName, endTime, endTime)[0][0]
            else:
                endTimeSQLVar = "endTime"
            sqlTimeCond.append("`%s`.`%s` <= %s" % (tableName, endTimeSQLVar, endTime))
        cmd += " WHERE %s" % " AND ".join(sqlTimeCond)
        # Calculate conditions
        sqlCondList = []
        for keyName in condDict:
            sqlORList = []
            if keyName in self.dbCatalog[typeName]["keys"]:
                List.appendUnique(
                    sqlLinkList, "`%s`.`%s` = `%s`.`id`" % (tableName, keyName, _getTableName("key", typeName, keyName))
                )
            if not isinstance(condDict[keyName], (list, tuple)):
                condDict[keyName] = [condDict[keyName]]
            for keyValue in condDict[keyName]:
                retVal = self._escapeString(keyValue)
                if not retVal["OK"]:
                    return retVal
                keyValue = retVal["Value"]
                if keyName in self.dbCatalog[typeName]["keys"]:
                    sqlORList.append("`%s`.`value` = %s" % (_getTableName("key", typeName, keyName), keyValue))
                else:
                    sqlORList.append("`%s`.`%s` = %s" % (tableName, keyName, keyValue))
            sqlCondList.append("( %s )" % " OR ".join(sqlORList))
        if sqlCondList:
            cmd += " AND %s" % " AND ".join(sqlCondList)
        # Calculate grouping and sorting
        for preGenFields in (groupFields, orderFields):
            if preGenFields:
                for i, field in enumerate(preGenFields[1]):
                    if field in self.dbCatalog[typeName]["keys"]:
                        List.appendUnique(
                            sqlLinkList,
                            "`%s`.`%s` = `%s`.`id`" % (tableName, field, _getTableName("key", typeName, field)),
                        )
                        if preGenFields[0] != "%s":
                            # The default grouping was changed
                            preGenFields[1][i] = "`%s`.Value" % _getTableName("key", typeName, field)
                        else:
                            # The default grouping is maintained
                            preGenFields[1][i] = "`%s`.`%s`" % (tableName, field)
                    elif field in ["bucketLength", "entriesInBucket"]:  # these are not in the dbCatalog
                        preGenFields[1][i] = "`%s`.`%s`" % (tableName, field)

        if sqlLinkList:
            cmd += " AND %s" % " AND ".join(sqlLinkList)
        if groupFields:
            if len(groupFields[1]) == 1:
                testGroupFields = " %s" % selectFields[0] % tuple(realFieldList)
                testGroupFieldsList = testGroupFields.split(",")
                realGroupFields = ()
                for testGroupFields in testGroupFieldsList:
                    if "sum" not in testGroupFields.lower():
                        realGroupFields += (testGroupFields.strip(),)
                cmd += " GROUP BY " + ",".join(realGroupFields)
            else:
                cmd += " GROUP BY %s" % (groupFields[0] % tuple(groupFields[1]))
        if orderFields:
            cmd += " ORDER BY %s" % (orderFields[0] % tuple(orderFields[1]))
        self.log.verbose(cmd)
        return self._query(cmd, conn=connObj)

    def compactBuckets(self, typeFilter=False):
        """
        Compact buckets for all defined types
        """
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        gSynchro.lock()
        try:
            if self.__doingCompaction:
                return S_OK()
            self.__doingCompaction = True
        finally:
            gSynchro.unlock()
        slow = True
        for typeName in self.dbCatalog:
            if typeFilter and typeName.find(typeFilter) == -1:
                self.log.info("[COMPACT] Skipping %s" % typeName)
                continue
            if self.dbCatalog[typeName]["dataTimespan"] > 0:
                self.log.info("[COMPACT] Deleting records older that timespan for type %s" % typeName)
                self.__deleteRecordsOlderThanDataTimespan(typeName)
            self.log.info("[COMPACT] Compacting %s" % typeName)
            if slow:
                self.__slowCompactBucketsForType(typeName)
            else:
                self.__compactBucketsForType(typeName)
        self.log.info("[COMPACT] Compaction finished")
        self.__lastCompactionEpoch = int(Time.toEpoch())
        gSynchro.lock()
        try:
            if self.__doingCompaction:
                self.__doingCompaction = False
        finally:
            gSynchro.unlock()
        return S_OK()

    def __selectForCompactBuckets(self, typeName, timeLimit, bucketLength, nextBucketLength, connObj=False):
        """
        Nasty SQL query to get ideal buckets using grouping by date calculations and adding value contents
        """
        tableName = _getTableName("bucket", typeName)
        selectSQL = "SELECT "
        sqlSelectList = []
        for field in self.dbCatalog[typeName]["keys"]:
            sqlSelectList.append("`%s`.`%s`" % (tableName, field))
        for field in self.dbCatalog[typeName]["values"]:
            sqlSelectList.append("SUM( `%s`.`%s` )" % (tableName, field))
        sqlSelectList.append("SUM( `%s`.`entriesInBucket` )" % (tableName))
        sqlSelectList.append("MIN( `%s`.`startTime` )" % tableName)
        sqlSelectList.append("MAX( `%s`.`startTime` )" % tableName)
        selectSQL += ", ".join(sqlSelectList)
        selectSQL += " FROM `%s`" % tableName
        selectSQL += " WHERE `%s`.`startTime` < '%s' AND" % (tableName, timeLimit)
        selectSQL += " `%s`.`bucketLength` = %s" % (tableName, bucketLength)
        # MAGIC bucketing
        sqlGroupList = [_bucketizeDataField("`%s`.`startTime`" % tableName, nextBucketLength)]
        for field in self.dbCatalog[typeName]["keys"]:
            sqlGroupList.append("`%s`.`%s`" % (tableName, field))
        selectSQL += " GROUP BY %s" % ", ".join(sqlGroupList)
        return self._query(selectSQL, conn=connObj)

    def __deleteForCompactBuckets(self, typeName, timeLimit, bucketLength, connObj=False):
        """
        Delete compacted buckets
        """
        tableName = _getTableName("bucket", typeName)
        deleteSQL = "DELETE FROM `%s` WHERE " % tableName
        deleteSQL += "`%s`.`startTime` < '%s' AND " % (tableName, timeLimit)
        deleteSQL += "`%s`.`bucketLength` = %s" % (tableName, bucketLength)
        return self._update(deleteSQL, conn=connObj)

    def __compactBucketsForType(self, typeName):
        """
        Compact all buckets for a given type
        """
        nowEpoch = Time.toEpoch()
        # retVal = self.__startTransaction(connObj)
        # if not retVal[ 'OK' ]:
        #  return retVal
        for bPos in range(len(self.dbBucketsLength[typeName]) - 1):
            self.log.info("[COMPACT] Query %d of %d" % (bPos + 1, len(self.dbBucketsLength[typeName]) - 1))
            secondsLimit = self.dbBucketsLength[typeName][bPos][0]
            bucketLength = self.dbBucketsLength[typeName][bPos][1]
            timeLimit = (nowEpoch - nowEpoch % bucketLength) - secondsLimit
            nextBucketLength = self.dbBucketsLength[typeName][bPos + 1][1]
            self.log.info(
                "[COMPACT] Compacting data newer that %s with bucket size %s"
                % (Time.fromEpoch(timeLimit), bucketLength)
            )
            # Retrieve the data
            retVal = self.__selectForCompactBuckets(typeName, timeLimit, bucketLength, nextBucketLength)
            if not retVal["OK"]:
                # self.__rollbackTransaction(connObj)
                return retVal
            bucketsData = retVal["Value"]
            self.log.info("[COMPACT] Got %d records to compact" % len(bucketsData))
            if len(bucketsData) == 0:
                continue
            retVal = self.__deleteForCompactBuckets(typeName, timeLimit, bucketLength)
            if not retVal["OK"]:
                # self.__rollbackTransaction( connObj )
                return retVal
            self.log.info(
                "[COMPACT] Compacting %s records %s seconds size for %s" % (len(bucketsData), bucketLength, typeName)
            )
            # Add data
            for record in bucketsData:
                startTime = record[-2]
                endTime = record[-1]
                valuesList = record[:-2]
                retVal = self.__splitInBuckets(typeName, startTime, endTime, valuesList)
                if not retVal["OK"]:
                    # self.__rollbackTransaction( connObj )
                    self.log.error(
                        "[COMPACT] Error while compacting data for record", "%s: %s" % (typeName, retVal["Value"])
                    )
            self.log.info("[COMPACT] Finished compaction %d of %d" % (bPos, len(self.dbBucketsLength[typeName]) - 1))
        # return self.__commitTransaction( connObj )
        return S_OK()

    def __slowCompactBucketsForType(self, typeName):
        """
        Compact all buckets for a given type
        """
        nowEpoch = Time.toEpoch()
        for bPos in range(len(self.dbBucketsLength[typeName]) - 1):
            self.log.info("[COMPACT] Query %d of %d" % (bPos, len(self.dbBucketsLength[typeName]) - 1))
            secondsLimit = self.dbBucketsLength[typeName][bPos][0]
            bucketLength = self.dbBucketsLength[typeName][bPos][1]
            timeLimit = (nowEpoch - nowEpoch % bucketLength) - secondsLimit
            self.log.info(
                "[COMPACT] Compacting data newer that %s with bucket size %s for %s"
                % (Time.fromEpoch(timeLimit), bucketLength, typeName)
            )
            querySize = 10000
            previousRecordsSelected = querySize
            totalCompacted = 0
            while previousRecordsSelected == querySize:
                # Retrieve the data
                self.log.info(
                    "[COMPACT] Retrieving buckets to compact newer that %s with size %s"
                    % (Time.fromEpoch(timeLimit), bucketLength)
                )
                roundStartTime = time.time()
                result = self.__selectIndividualForCompactBuckets(typeName, timeLimit, bucketLength, querySize)
                if not result["OK"]:
                    # self.__rollbackTransaction( connObj )
                    return result
                bucketsData = result["Value"]
                previousRecordsSelected = len(bucketsData)
                selectEndTime = time.time()
                self.log.info(
                    "[COMPACT] Got %d buckets (%d done) (took %.2f secs)"
                    % (previousRecordsSelected, totalCompacted, selectEndTime - roundStartTime)
                )
                if len(bucketsData) == 0:
                    break

                result = self.__deleteIndividualForCompactBuckets(typeName, bucketsData)
                if not result["OK"]:
                    # self.__rollbackTransaction(connObj)
                    return result
                bucketsData = result["Value"]
                deleteEndTime = time.time()
                self.log.info(
                    "[COMPACT] Deleted %s out-of-bounds buckets (took %.2f secs)"
                    % (len(bucketsData), deleteEndTime - selectEndTime)
                )
                # Add data
                for record in bucketsData:
                    startTime = record[-2]
                    endTime = record[-2] + record[-1]
                    valuesList = record[:-2]
                    retVal = self.__splitInBuckets(typeName, startTime, endTime, valuesList)
                    if not retVal["OK"]:
                        self.log.error(
                            "[COMPACT] Error while compacting data for buckets", "%s: %s" % (typeName, retVal["Value"])
                        )
                totalCompacted += len(bucketsData)
                insertElapsedTime = time.time() - deleteEndTime
                self.log.info(
                    "[COMPACT] Records compacted (took %.2f secs, %.2f secs/bucket)"
                    % (insertElapsedTime, insertElapsedTime / len(bucketsData))
                )
            self.log.info("[COMPACT] Finised compaction %d of %d" % (bPos, len(self.dbBucketsLength[typeName]) - 1))
        # return self.__commitTransaction(connObj)
        return S_OK()

    def __selectIndividualForCompactBuckets(self, typeName, timeLimit, bucketLength, querySize, connObj=False):
        """
        Nasty SQL query to get ideal buckets using grouping by date calculations and adding value contents
        """
        tableName = _getTableName("bucket", typeName)
        selectSQL = "SELECT "
        sqlSelectList = []
        for field in self.dbCatalog[typeName]["keys"]:
            sqlSelectList.append("`%s`.`%s`" % (tableName, field))
        for field in self.dbCatalog[typeName]["values"]:
            sqlSelectList.append("`%s`.`%s`" % (tableName, field))
        sqlSelectList.append("`%s`.`entriesInBucket`" % (tableName))
        sqlSelectList.append("`%s`.`startTime`" % tableName)
        sqlSelectList.append("`%s`.bucketLength" % (tableName))
        selectSQL += ", ".join(sqlSelectList)
        selectSQL += " FROM `%s`" % tableName
        selectSQL += " WHERE `%s`.`startTime` < '%s' AND" % (tableName, timeLimit)
        selectSQL += " `%s`.`bucketLength` = %s" % (tableName, bucketLength)
        # MAGIC bucketing
        selectSQL += " LIMIT %d" % querySize
        return self._query(selectSQL, conn=connObj)

    def __deleteIndividualForCompactBuckets(self, typeName, bucketsData, connObj=False):
        """
        Delete compacted buckets
        """
        tableName = _getTableName("bucket", typeName)
        keyFields = self.dbCatalog[typeName]["keys"]
        deleteQueryLimit = 50
        deletedBuckets = []
        for bLimit in range(0, len(bucketsData), deleteQueryLimit):
            delCondsSQL = []
            for record in bucketsData[bLimit : bLimit + deleteQueryLimit]:
                condSQL = []
                for iPos, field in enumerate(keyFields):
                    condSQL.append("`%s`.`%s` = %s" % (tableName, field, record[iPos]))
                condSQL.append("`%s`.`startTime` = %d" % (tableName, record[-2]))
                condSQL.append("`%s`.`bucketLength` = %d" % (tableName, record[-1]))
                delCondsSQL.append("(%s)" % " AND ".join(condSQL))
            delSQL = "DELETE FROM `%s` WHERE %s" % (tableName, " OR ".join(delCondsSQL))
            result = self._update(delSQL, conn=connObj)
            if not result["OK"]:
                self.log.error("Cannot delete individual records for compaction", result["Message"])
            else:
                deletedBuckets.extend(bucketsData[bLimit : bLimit + deleteQueryLimit])
        return S_OK(deletedBuckets)

    def __deleteRecordsOlderThanDataTimespan(self, typeName):
        """
        IF types define dataTimespan, then records older than datatimespan seconds will be deleted
        automatically
        """
        dataTimespan = self.dbCatalog[typeName]["dataTimespan"] + self.dbBucketsLength[typeName][-1][1]
        if dataTimespan < 86400 * 30:
            return
        for table, field in (
            (_getTableName("type", typeName), "endTime"),
            (_getTableName("bucket", typeName), "startTime"),
        ):
            self.log.info("[COMPACT] Deleting old records for table %s" % table)
            deleteLimit = 100000
            deleted = deleteLimit
            while deleted >= deleteLimit:
                sqlCmd = "DELETE FROM `%s` WHERE %s < UNIX_TIMESTAMP()-%d LIMIT %d" % (
                    table,
                    field,
                    dataTimespan,
                    deleteLimit,
                )
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error(
                        "[COMPACT] Cannot delete old records",
                        "Table: %s Timespan: %s Error: %s" % (table, dataTimespan, result["Message"]),
                    )
                    break
                self.log.info("[COMPACT] Deleted %d records for %s table" % (result["Value"], table))
                deleted = result["Value"]
                time.sleep(1)

    def regenerateBuckets(self, typeName):
        if self.__readOnly:
            return S_ERROR("ReadOnly mode enabled. No modification allowed")
        # Delete old entries if any
        if self.dbCatalog[typeName]["dataTimespan"] > 0:
            self.log.info("[REBUCKET] Deleting records older that timespan for type %s" % typeName)
            self.__deleteRecordsOlderThanDataTimespan(typeName)
            self.log.info("[REBUCKET] Done deleting old records")
        rawTableName = _getTableName("type", typeName)
        # retVal = self.__startTransaction(connObj)
        # if not retVal[ 'OK' ]:
        #  return retVal
        self.log.info("[REBUCKET] Deleting buckets for %s" % typeName)
        retVal = self._update("DELETE FROM `%s`" % _getTableName("bucket", typeName))
        if not retVal["OK"]:
            return retVal
        # Generate the common part of the query
        # SELECT fields
        startTimeTableField = "`%s`.startTime" % rawTableName
        endTimeTableField = "`%s`.endTime" % rawTableName
        # Select strings and sum select strings
        sqlSUMSelectList = []
        sqlSelectList = []
        for field in self.dbCatalog[typeName]["keys"]:
            sqlSUMSelectList.append("`%s`.`%s`" % (rawTableName, field))
            sqlSelectList.append("`%s`.`%s`" % (rawTableName, field))
        for field in self.dbCatalog[typeName]["values"]:
            sqlSUMSelectList.append("SUM( `%s`.`%s` )" % (rawTableName, field))
            sqlSelectList.append("`%s`.`%s`" % (rawTableName, field))
        sumSelectString = ", ".join(sqlSUMSelectList)
        selectString = ", ".join(sqlSelectList)
        # Grouping fields
        sqlGroupList = []
        for field in self.dbCatalog[typeName]["keys"]:
            sqlGroupList.append("`%s`.`%s`" % (rawTableName, field))
        groupingString = ", ".join(sqlGroupList)
        # List to contain all queries
        sqlQueries = []
        dateInclusiveConditions = []
        countedField = "`%s`.`%s`" % (rawTableName, self.dbCatalog[typeName]["keys"][0])
        lastTime = Time.toEpoch()
        # Iterate for all ranges
        for iRange, iValue in enumerate(self.dbBucketsLength[typeName]):
            bucketTimeSpan = iValue[0]
            bucketLength = iValue[1]
            startRangeTime = lastTime - bucketTimeSpan
            endRangeTime = lastTime
            lastTime -= bucketTimeSpan
            bucketizedStart = _bucketizeDataField(startTimeTableField, bucketLength)
            bucketizedEnd = _bucketizeDataField(endTimeTableField, bucketLength)

            timeSelectString = "MIN(%s), MAX(%s)" % (startTimeTableField, endTimeTableField)
            # Is the last bucket?
            if iRange == len(self.dbBucketsLength[typeName]) - 1:
                whereString = "%s <= %d" % (endTimeTableField, endRangeTime)
            else:
                whereString = "%s > %d AND %s <= %d" % (
                    startTimeTableField,
                    startRangeTime,
                    endTimeTableField,
                    endRangeTime,
                )
            sameBucketCondition = "(%s) = (%s)" % (bucketizedStart, bucketizedEnd)
            # Records that fit in a bucket
            sqlQuery = "SELECT %s, %s, COUNT(%s) FROM `%s` WHERE %s AND %s GROUP BY %s, %s" % (
                timeSelectString,
                sumSelectString,
                countedField,
                rawTableName,
                whereString,
                sameBucketCondition,
                groupingString,
                bucketizedStart,
            )
            sqlQueries.append(sqlQuery)
            # Records that fit in more than one bucket
            sqlQuery = "SELECT %s, %s, %s, 1 FROM `%s` WHERE %s AND NOT %s" % (
                startTimeTableField,
                endTimeTableField,
                selectString,
                rawTableName,
                whereString,
                sameBucketCondition,
            )
            sqlQueries.append(sqlQuery)
            dateInclusiveConditions.append("( %s )" % whereString)
        # Query for records that are in between two ranges
        sqlQuery = "SELECT %s, %s, %s, 1 FROM `%s` WHERE NOT %s" % (
            startTimeTableField,
            endTimeTableField,
            selectString,
            rawTableName,
            " AND NOT ".join(dateInclusiveConditions),
        )
        sqlQueries.append(sqlQuery)
        self.log.info("[REBUCKET] Retrieving data for rebuilding buckets for type %s..." % (typeName))
        queryNum = 0
        for sqlQuery in sqlQueries:
            self.log.info("[REBUCKET] Executing query #%s..." % queryNum)
            queryNum += 1
            retVal = self._query(sqlQuery)
            if not retVal["OK"]:
                self.log.error("[REBUCKET] Can't retrieve data for rebucketing", retVal["Message"])
                # self.__rollbackTransaction(connObj)
                return retVal
            rawData = retVal["Value"]
            self.log.info("[REBUCKET] Retrieved %s records" % len(rawData))
            rebucketedRecords = 0
            startQuery = time.time()
            startBlock = time.time()
            numRecords = len(rawData)
            for entry in rawData:
                startT = entry[0]
                endT = entry[1]
                values = entry[2:]
                retVal = self.__splitInBuckets(typeName, startT, endT, values)
                if not retVal["OK"]:
                    # self.__rollbackTransaction(connObj)
                    return retVal
                rebucketedRecords += 1
                if rebucketedRecords % 1000 == 0:
                    queryAvg = rebucketedRecords / float(time.time() - startQuery)
                    blockAvg = 1000 / float(time.time() - startBlock)
                    startBlock = time.time()
                    perDone = 100 * rebucketedRecords / float(numRecords)
                    expectedEnd = str(datetime.timedelta(seconds=int((numRecords - rebucketedRecords) / blockAvg)))
                    self.log.info(
                        "[REBUCKET] Rebucketed %.2f%% %s (%.2f r/s block %.2f r/s query | ETA %s )..."
                        % (perDone, typeName, blockAvg, queryAvg, expectedEnd)
                    )
        # return self.__commitTransaction(connObj)
        return S_OK()

    def __startTransaction(self, connObj):
        return self._query("START TRANSACTION", conn=connObj)

    def __commitTransaction(self, connObj):
        return self._query("COMMIT", conn=connObj)

    def __rollbackTransaction(self, connObj):
        return self._query("ROLLBACK", conn=connObj)


def _bucketizeDataField(dataField, bucketLength):
    return "%s - ( %s %% %s )" % (dataField, dataField, bucketLength)


def _getTableName(tableType, typeName, keyName=None):
    """
    Generate table name
    """
    if not keyName:
        return "ac_%s_%s" % (tableType, typeName)
    elif tableType == "key":
        return "ac_%s_%s_%s" % (tableType, typeName, keyName)
    else:
        raise Exception("Call to _getTableName with tableType as key but with no keyName")



########################################################################
# File :    SiteDirector.py
# Author :  A.T., F.S.
########################################################################

"""  The Site Director is an agent performing pilot job submission to particular sites/Computing Elements.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN SiteDirector
  :end-before: ##END
  :dedent: 2
  :caption: SiteDirector options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import sys
import random
import socket
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor

import DIRAC
from DIRAC import S_OK, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals, Registry
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.Time import dateTime, second
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.AccountingSystem.Client.Types.Pilot import Pilot as PilotAccounting
from DIRAC.AccountingSystem.Client.Types.PilotSubmission import PilotSubmission as PilotSubmissionAccounting
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.WorkloadManagementSystem.Client import PilotStatus
from DIRAC.WorkloadManagementSystem.Client.MatcherClient import MatcherClient
from DIRAC.WorkloadManagementSystem.Client.ServerUtils import pilotAgentsDB
from DIRAC.WorkloadManagementSystem.Service.WMSUtilities import getGridEnv
from DIRAC.WorkloadManagementSystem.private.ConfigHelper import findGenericPilotCredentials
from DIRAC.WorkloadManagementSystem.Utilities.QueueUtilities import getQueuesResolved
from DIRAC.WorkloadManagementSystem.Utilities.PilotWrapper import (
    pilotWrapperScript,
    _writePilotWrapperFile,
    getPilotFilesCompressedEncodedDict,
)
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus

# dirac install file
DIRAC_INSTALL = os.path.join("Core", "scripts", "dirac-install.py")

MAX_PILOTS_TO_SUBMIT = 100


class SiteDirector(AgentModule):
    """SiteDirector class provides an implementation of a DIRAC agent.

    Used for submitting pilots to Computing Elements.
    """

    def __init__(self, *args, **kwargs):
        """c'tor"""
        super(SiteDirector, self).__init__(*args, **kwargs)

        # on-the fly imports
        ol = ObjectLoader()
        res = ol.loadModule("ConfigurationSystem.Client.Helpers.Resources")
        if not res["OK"]:
            sys.exit(res["Message"])
        self.resourcesModule = res["Value"]

        self.queueDict = {}
        # self.queueCECache aims at saving CEs information over the cycles to avoid to create the exact same CEs each cycle
        self.queueCECache = {}
        self.queueSlots = {}
        self.failedQueues = defaultdict(int)
        # failedPilotOutput stores the number of times the Site Director failed to get a given pilot output
        self.failedPilotOutput = defaultdict(int)
        self.firstPass = True
        self.maxPilotsToSubmit = MAX_PILOTS_TO_SUBMIT

        self.gridEnv = ""
        self.vo = ""
        self.group = ""
        # self.voGroups contain all the eligible user groups for pilots submitted by this SiteDirector
        self.voGroups = []
        self.pilotDN = ""
        self.pilotGroup = ""
        self.platforms = []
        self.sites = []
        self.totalSubmittedPilots = 0

        self.python3Pilots = True
        self.addPilotsToEmptySites = False
        self.checkPlatform = False
        self.updateStatus = True
        self.getOutput = False
        self.sendAccounting = True
        self.sendSubmissionAccounting = True

        self.siteClient = None
        self.rssClient = None
        self.rssFlag = None

        self.globalParameters = {"NumberOfProcessors": 1, "MaxRAM": 2048}
        # self.failedQueueCycleFactor is the number of cycles a queue has to wait before getting pilots again
        self.failedQueueCycleFactor = 10
        # Every N cycles, the status of the pilots are updated by the SiteDirector
        self.pilotStatusUpdateCycleFactor = 10
        # Every N cycles, the number of slots available in the queues is updated
        self.availableSlotsUpdateCycleFactor = 10
        self.maxQueueLength = 86400 * 3
        # Maximum number of times the Site Director is going to try to get a pilot output before stopping
        self.maxRetryGetPilotOutput = 3

        self.pilotWaitingFlag = True
        self.pilotLogLevel = "INFO"
        self.matcherClient = None
        self.siteMaskList = []
        self.ceMaskList = []

        self.localhost = socket.getfqdn()

    def initialize(self):
        """Initial settings"""

        self.gridEnv = self.am_getOption("GridEnv", "")
        if not self.gridEnv:
            self.gridEnv = getGridEnv()

        # The SiteDirector is for a particular user community
        self.vo = self.am_getOption("VO", "")
        if not self.vo:
            self.vo = self.am_getOption("Community", "")
        if not self.vo:
            self.vo = CSGlobals.getVO()
        # The SiteDirector is for a particular user group
        self.group = self.am_getOption("Group", "")

        # Choose the group for which pilots will be submitted. This is a hack until
        # we will be able to match pilots to VOs.
        if not self.group:
            if self.vo:
                result = Registry.getGroupsForVO(self.vo)
                if not result["OK"]:
                    return result
                self.voGroups = []
                for group in result["Value"]:
                    if "NormalUser" in Registry.getPropertiesForGroup(group):
                        self.voGroups.append(group)
        else:
            self.voGroups = [self.group]

        # Get the clients
        self.siteClient = SiteStatus()
        self.rssClient = ResourceStatus()
        self.matcherClient = MatcherClient()

        return S_OK()

    def beginExecution(self):
        """This is run at every cycle, as first thing.

        1. Check the pilots credentials.
        2. Get some flags and options used later
        3. Get the site description dictionary
        4. Get what to send in pilot wrapper
        """

        self.rssFlag = self.rssClient.rssFlag

        # Which credentials to use?
        # are they specific to the SD? (if not, get the generic ones)
        self.pilotDN = self.am_getOption("PilotDN", self.pilotDN)
        self.pilotGroup = self.am_getOption("PilotGroup", self.pilotGroup)
        result = findGenericPilotCredentials(vo=self.vo, pilotDN=self.pilotDN, pilotGroup=self.pilotGroup)
        if not result["OK"]:
            return result
        self.pilotDN, self.pilotGroup = result["Value"]

        # Parameters
        self.workingDirectory = self.am_getOption("WorkDirectory")
        self.maxQueueLength = self.am_getOption("MaxQueueLength", self.maxQueueLength)
        self.pilotLogLevel = self.am_getOption("PilotLogLevel", self.pilotLogLevel)
        self.maxPilotsToSubmit = self.am_getOption("MaxPilotsToSubmit", self.maxPilotsToSubmit)
        self.pilotWaitingFlag = self.am_getOption("PilotWaitingFlag", self.pilotWaitingFlag)
        self.failedQueueCycleFactor = self.am_getOption("FailedQueueCycleFactor", self.failedQueueCycleFactor)
        self.pilotStatusUpdateCycleFactor = self.am_getOption(
            "PilotStatusUpdateCycleFactor", self.pilotStatusUpdateCycleFactor
        )
        self.availableSlotsUpdateCycleFactor = self.am_getOption(
            "AvailableSlotsUpdateCycleFactor", self.availableSlotsUpdateCycleFactor
        )
        self.maxRetryGetPilotOutput = self.am_getOption("MaxRetryGetPilotOutput", self.maxRetryGetPilotOutput)

        # Flags
        self.python3Pilots = self.am_getOption("Python3Pilots", self.python3Pilots)
        self.addPilotsToEmptySites = self.am_getOption("AddPilotsToEmptySites", self.addPilotsToEmptySites)
        self.checkPlatform = self.am_getOption("CheckPlatform", self.checkPlatform)
        self.updateStatus = self.am_getOption("UpdatePilotStatus", self.updateStatus)
        self.getOutput = self.am_getOption("GetPilotOutput", self.getOutput)
        self.sendAccounting = self.am_getOption("SendPilotAccounting", self.sendAccounting)
        self.sendSubmissionAccounting = self.am_getOption(
            "SendPilotSubmissionAccounting", self.sendSubmissionAccounting
        )

        # Get the site description dictionary
        siteNames = None
        siteNamesOption = self.am_getOption("Site", ["any"])
        if siteNamesOption and "any" not in [sn.lower() for sn in siteNamesOption]:
            siteNames = siteNamesOption

        ceTypes = None
        ceTypesOption = self.am_getOption("CETypes", ["any"])
        if ceTypesOption and "any" not in [ct.lower() for ct in ceTypesOption]:
            ceTypes = ceTypesOption

        ces = None
        cesOption = self.am_getOption("CEs", ["any"])
        if cesOption and "any" not in [ce.lower() for ce in cesOption]:
            ces = cesOption

        self.log.always("VO:", self.vo)
        if self.voGroups:
            self.log.always("Group(s):", self.voGroups)
        self.log.always("Sites:", siteNames)
        self.log.always("CETypes:", ceTypes)
        self.log.always("CEs:", ces)
        self.log.always("PilotDN:", self.pilotDN)
        self.log.always("PilotGroup:", self.pilotGroup)

        result = self.resourcesModule.getQueues(
            community=self.vo, siteList=siteNames, ceList=ces, ceTypeList=ceTypes, mode="Direct"
        )
        if not result["OK"]:
            return result
        result = getQueuesResolved(
            siteDict=result["Value"],
            queueCECache=self.queueCECache,
            gridEnv=self.gridEnv,
            setup=gConfig.getValue("/DIRAC/Setup", "unknown"),
            workingDir=self.workingDirectory,
            checkPlatform=self.checkPlatform,
            instantiateCEs=True,
        )
        if not result["OK"]:
            return result

        self.queueDict = result["Value"]
        for queueName, queueDict in self.queueDict.items():

            # Update self.sites
            if queueDict["Site"] not in self.sites:
                self.sites.append(queueDict["Site"])

            # Update self.platforms, keeping entries unique and squashing lists
            self.platforms = []
            if "Platform" in queueDict["ParametersDict"]:
                platform = queueDict["ParametersDict"]["Platform"]
                oldPlatforms = set(self.platforms)
                if isinstance(platform, list):
                    oldPlatforms.update(set(platform))
                else:
                    oldPlatforms.add(platform)
                self.platforms = list(oldPlatforms)

            # Update self.globalParameters
            if "WholeNode" in queueDict["ParametersDict"]:
                self.globalParameters["WholeNode"] = "True"
            for parameter in ["MaxRAM", "NumberOfProcessors"]:
                if parameter in queueDict["ParametersDict"]:
                    self.globalParameters[parameter] = max(
                        self.globalParameters[parameter], int(queueDict["ParametersDict"][parameter])
                    )

        if self.updateStatus:
            self.log.always("Pilot status update requested")
        if self.getOutput:
            self.log.always("Pilot output retrieval requested")
        if self.sendAccounting:
            self.log.always("Pilot accounting sending requested")
        if self.sendSubmissionAccounting:
            self.log.always("Pilot submission accounting sending requested")

        self.log.always("MaxPilotsToSubmit:", self.maxPilotsToSubmit)

        if self.firstPass:
            if self.queueDict:
                self.log.always("Agent will serve queues:")
                for queue in self.queueDict:
                    self.log.always(
                        "Site: %s, CE: %s, Queue: %s"
                        % (self.queueDict[queue]["Site"], self.queueDict[queue]["CEName"], queue)
                    )
        self.firstPass = False

        return S_OK()

    def execute(self):
        """Main execution method (what is called at each agent cycle).

        It basically just calls self.submitPilots() method
        """

        if not self.queueDict:
            self.log.warn("No site defined, exiting the cycle")
            return S_OK()

        # get list of usable sites within this cycle
        result = self.siteClient.getUsableSites()
        if not result["OK"]:
            return result
        self.siteMaskList = result.get("Value", [])

        if self.rssFlag:
            ceNamesList = [queue["CEName"] for queue in self.queueDict.values()]
            result = self.rssClient.getElementStatus(ceNamesList, "ComputingElement", vO=self.vo)
            if not result["OK"]:
                self.log.error("Can not get the status of computing elements", " %s" % result["Message"])
                return result
            # Try to get CEs which have been probed and those unprobed (vO='all').
            self.ceMaskList = [
                ceName for ceName in result["Value"] if result["Value"][ceName]["all"] in ("Active", "Degraded")
            ]
            self.log.debug("CE list with status Active or Degraded: ", self.ceMaskList)

        result = self.submitPilots()
        if not result["OK"]:
            self.log.error("Errors in the job submission: ", result["Message"])
            return result

        # Every N cycles we update the pilots status
        cyclesDone = self.am_getModuleParam("cyclesDone")
        if self.updateStatus and cyclesDone % self.pilotStatusUpdateCycleFactor == 0:
            result = self.updatePilotStatus()
            if not result["OK"]:
                self.log.error("Errors in updating pilot status: ", result["Message"])
                return result

        return S_OK()

    def submitPilots(self):
        """Go through defined computing elements and submit pilots if necessary and possible

        :return: S_OK/S_ERROR
        """

        # First, we check if we want to submit pilots at all, and also where
        submit, anySite, jobSites, testSites = self._ifAndWhereToSubmit()
        if not submit:
            self.log.notice("Not submitting any pilots at this cycle")
            return S_OK()

        # From here on we assume we are going to (try to) submit some pilots
        self.log.debug("Going to try to submit some pilots")

        self.log.verbose("Queues treated", ",".join(self.queueDict))

        self.totalSubmittedPilots = 0

        queueDictItems = list(self.queueDict.items())
        random.shuffle(queueDictItems)

        for queueName, queueDictionary in queueDictItems:
            # now submitting to the single queues
            self.log.verbose("Evaluating queue", queueName)

            # are we going to submit pilots to this specific queue?
            if not self._allowedToSubmit(queueName, anySite, jobSites, testSites):
                continue

            if "CPUTime" in queueDictionary["ParametersDict"]:
                queueCPUTime = int(queueDictionary["ParametersDict"]["CPUTime"])
            else:
                self.log.warn("CPU time limit is not specified, skipping", "queue %s" % queueName)
                continue
            if queueCPUTime > self.maxQueueLength:
                queueCPUTime = self.maxQueueLength

            ce, ceDict = self._getCE(queueName)

            # additionalInfo is normally taskQueueDict
            pilotsWeMayWantToSubmit, additionalInfo = self._getPilotsWeMayWantToSubmit(ceDict)
            self.log.debug(
                "%d pilotsWeMayWantToSubmit are eligible for %s queue" % (pilotsWeMayWantToSubmit, queueName)
            )
            if not pilotsWeMayWantToSubmit:
                self.log.debug("...so skipping %s" % queueName)
                continue

            # Get the number of already waiting pilots for the queue
            totalWaitingPilots = 0
            manyWaitingPilotsFlag = False
            if self.pilotWaitingFlag:
                tqIDList = list(additionalInfo)
                result = pilotAgentsDB.countPilots(
                    {"TaskQueueID": tqIDList, "Status": PilotStatus.PILOT_WAITING_STATES}, None
                )
                if not result["OK"]:
                    self.log.error("Failed to get Number of Waiting pilots", result["Message"])
                    totalWaitingPilots = 0
                else:
                    totalWaitingPilots = result["Value"]
                    self.log.debug("Waiting Pilots: %s" % totalWaitingPilots)
            if totalWaitingPilots >= pilotsWeMayWantToSubmit:
                self.log.verbose("Possibly enough pilots already waiting", "(%d)" % totalWaitingPilots)
                manyWaitingPilotsFlag = True
                if not self.addPilotsToEmptySites:
                    continue

            self.log.debug(
                "%d waiting pilots for the total of %d eligible pilots for %s"
                % (totalWaitingPilots, pilotsWeMayWantToSubmit, queueName)
            )

            # Get the number of available slots on the target site/queue
            totalSlots = self.getQueueSlots(queueName, manyWaitingPilotsFlag)
            if totalSlots <= 0:
                self.log.debug("%s: No slots available" % queueName)
                continue

            if manyWaitingPilotsFlag:
                # Throttle submission of extra pilots to empty sites
                pilotsToSubmit = int(self.maxPilotsToSubmit / 10) + 1
            else:
                pilotsToSubmit = max(0, min(totalSlots, pilotsWeMayWantToSubmit - totalWaitingPilots))
                self.log.info(
                    "%s: Slots=%d, TQ jobs(pilotsWeMayWantToSubmit)=%d, Pilots: waiting %d, to submit=%d"
                    % (queueName, totalSlots, pilotsWeMayWantToSubmit, totalWaitingPilots, pilotsToSubmit)
                )

            # Limit the number of pilots to submit to MAX_PILOTS_TO_SUBMIT
            pilotsToSubmit = min(self.maxPilotsToSubmit, pilotsToSubmit)

            # Get the working proxy
            cpuTime = queueCPUTime + 86400
            self.log.verbose("Getting pilot proxy", "for %s/%s %d long" % (self.pilotDN, self.pilotGroup, cpuTime))
            result = gProxyManager.getPilotProxyFromDIRACGroup(self.pilotDN, self.pilotGroup, cpuTime)
            if not result["OK"]:
                return result
            proxy = result["Value"]
            # Check returned proxy lifetime
            result = proxy.getRemainingSecs()  # pylint: disable=no-member
            if not result["OK"]:
                return result
            lifetime_secs = result["Value"]
            ce.setProxy(proxy, lifetime_secs)

            # now really submitting
            res = self._submitPilotsToQueue(pilotsToSubmit, ce, queueName)
            if not res["OK"]:
                self.log.info("Failed pilot submission", "Queue: %s" % queueName)
            else:
                pilotList, stampDict = res["Value"]

                # updating the pilotAgentsDB... done by default but maybe not strictly necessary
                self._addPilotTQReference(queueName, additionalInfo, pilotList, stampDict)

        # Summary after the cycle over queues
        self.log.info("Total number of pilots submitted in this cycle", "%d" % self.totalSubmittedPilots)

        return S_OK()

    def _ifAndWhereToSubmit(self):
        """Return a tuple that says if and where to submit pilots:

        (submit, anySite, jobSites, testSites)
        e.g.
        (True, False, {'Site1', 'Site2'}, {'Test1', 'Test2'})

        VOs may want to replace this method with different strategies
        """

        tqDict = self._getTQDictForMatching()
        if not tqDict:
            return True, True, set(), set()

        # the tqDict used here is a very generic one, not specific to one CE/queue only
        self.log.verbose("Checking overall TQ availability with requirements")
        self.log.verbose(tqDict)

        # Check that there is some work at all
        result = self.matcherClient.getMatchingTaskQueues(tqDict)
        if not result["OK"]:
            self.log.error("Matcher error:", result["Message"])
            return False, True, set(), set()
        matchingTQs = result["Value"]
        if not matchingTQs:
            self.log.notice("No Waiting jobs suitable for the director, so nothing to submit")
            return False, True, set(), set()

        # If we are here there's some work to do, now let's see for where
        jobSites = set()
        testSites = set()
        anySite = False

        for tqDescription in matchingTQs.values():
            siteList = tqDescription.get("Sites", [])
            if siteList:
                jobSites |= set(siteList)
            else:
                anySite = True

            if "JobTypes" in tqDescription:
                if "Sites" in tqDescription:
                    for site in tqDescription["Sites"]:
                        if site.lower() != "any":
                            testSites.add(site)

        self.monitorJobsQueuesPilots(matchingTQs)

        return True, anySite, jobSites, testSites

    def monitorJobsQueuesPilots(self, matchingTQs):
        """Just printout of jobs queues and pilots status in TQ"""
        tqIDList = list(matchingTQs)
        result = pilotAgentsDB.countPilots({"TaskQueueID": tqIDList, "Status": PilotStatus.PILOT_WAITING_STATES}, None)

        totalWaitingJobs = 0
        for tqDescription in matchingTQs.values():
            totalWaitingJobs += tqDescription["Jobs"]

        if not result["OK"]:
            self.log.error("Can't count pilots", result["Message"])
        else:
            self.log.info(
                "Total jobs : number of task queues : number of waiting pilots",
                "%d : %d : %d" % (totalWaitingJobs, len(tqIDList), result["Value"]),
            )

    def _getTQDictForMatching(self):
        """Just construct a dictionary (tqDict)
        that will be used to check with Matcher if there's anything to submit.

        If extensions want, they can replace partly or fully this method.
        If it returns just an empty dict, the assuption is that we'll submit pilots no matters what.

        :returns dict: tqDict of task queue descriptions
        """
        tqDict = {"Setup": CSGlobals.getSetup(), "CPUTime": 9999999}
        if self.vo:
            tqDict["Community"] = self.vo
        if self.voGroups:
            tqDict["OwnerGroup"] = self.voGroups

        if self.checkPlatform:
            platforms = self._getPlatforms()
            if platforms:
                tqDict["Platform"] = platforms

        tqDict["Site"] = self.sites

        # Get a union of all tags
        tags = []
        for queue in self.queueDict:
            tags += self.queueDict[queue]["ParametersDict"].get("Tag", [])
        tqDict["Tag"] = list(set(tags))

        # Add overall max values for all queues
        tqDict.update(self.globalParameters)

        return tqDict

    def _getPlatforms(self):
        """Get the platforms used for TQ match
        Here for extension purpose.

        :return: list of platforms
        """
        result = self.resourcesModule.getCompatiblePlatforms(self.platforms)
        if not result["OK"]:
            self.log.error(
                "Issue getting compatible platforms, will skip check of platforms",
                self.platforms + " : " + result["Message"],
            )
        return result["Value"]

    def _allowedToSubmit(self, queue, anySite, jobSites, testSites):
        """Check if we are allowed to submit to a certain queue

        :param str queue: the queue name
        :param bool anySite: submitting anywhere?
        :param set jobSites: set of job site names (only considered if anySite is False)
        :param set testSites: set of test site names

        :return: True/False
        """

        # Check if the queue failed previously
        failedCount = self.failedQueues[queue] % self.failedQueueCycleFactor
        if failedCount != 0:
            self.log.warn("queue failed recently ==> number of cycles skipped", "%s ==> %d" % (queue, 10 - failedCount))
            self.failedQueues[queue] += 1
            return False

        # Check the status of the site
        if self.queueDict[queue]["Site"] not in self.siteMaskList and self.queueDict[queue]["Site"] not in testSites:
            self.log.verbose(
                "Queue skipped (site not in mask)",
                "%s (%s)" % (self.queueDict[queue]["QueueName"], self.queueDict[queue]["Site"]),
            )
            return False

        # Check that there are task queues waiting for this site
        if not anySite and self.queueDict[queue]["Site"] not in jobSites:
            self.log.verbose(
                "Queue skipped: no workload expected",
                "%s at %s" % (self.queueDict[queue]["QueueName"], self.queueDict[queue]["Site"]),
            )
            return False

        # Check the status of the CE (only for RSS=Active)
        if self.rssFlag:
            if self.queueDict[queue]["CEName"] not in self.ceMaskList:
                self.log.verbose(
                    "Skipping computing element: resource not usable",
                    "%s at %s" % (self.queueDict[queue]["CEName"], self.queueDict[queue]["Site"]),
                )
                return False

        # if we are here, it means that we are allowed to submit to the queue
        return True

    def _getCE(self, queue):
        """Prepare the queue description to look for eligible jobs

        :param str queue: queue name

        :return: ce (ComputingElement object), ceDict (dict)
        """

        ce = self.queueDict[queue]["CE"]
        ceDict = ce.ceParameters
        ceDict["GridCE"] = self.queueDict[queue]["CEName"]

        if self.queueDict[queue]["Site"] not in self.siteMaskList:
            ceDict["JobType"] = "Test"
        if self.vo:
            ceDict["Community"] = self.vo
        if self.voGroups:
            ceDict["OwnerGroup"] = self.voGroups

        if self.checkPlatform:
            result = self.resourcesModule.getCompatiblePlatforms(self.queueDict[queue]["Platform"])
            if not result["OK"]:
                self.log.error(
                    "Issue getting compatible platforms, returning 'ANY'",
                    "%s: %s" % (self.platforms, result["Message"]),
                )
                ceDict["Platform"] = "ANY"
            ceDict["Platform"] = result["Value"]

        return ce, ceDict

    def _getPilotsWeMayWantToSubmit(self, ceDict):
        """Returns the number of pilots that we may want to submit to the ce described in ceDict

        This implementation is based on the number of eligible WMS taskQueues for the target site/queue.
        VOs are free to override this method and to provide a different implementation.

        :param ceDict: dictionary describing CE
        :type ceDict: dict

        :return: pilotsWeMayWantToSubmit (int), taskQueueDict (dict)
        :rType: tuple
        """

        pilotsWeMayWantToSubmit = 0

        result = self.matcherClient.getMatchingTaskQueues(ceDict)
        if not result["OK"]:
            self.log.error("Could not retrieve TaskQueues from TaskQueueDB", result["Message"])
            return 0, {}
        taskQueueDict = result["Value"]
        if not taskQueueDict:
            self.log.verbose("No matching TQs found", "for %s" % ceDict)

        for tq in taskQueueDict.values():
            pilotsWeMayWantToSubmit += tq["Jobs"]

        return pilotsWeMayWantToSubmit, taskQueueDict

    def _submitPilotsToQueue(self, pilotsToSubmit, ce, queue):
        """Method that really submits the pilots to the ComputingElements' queue

        :param pilotsToSubmit: number of pilots to submit.
        :type pilotsToSubmit: int
        :param ce: computing element object to where we submit
        :type ce: ComputingElement
        :param str queue: queue where to submit

        :return: S_OK/S_ERROR.
                 If S_OK, returns tuple with (pilotsToSubmit, pilotList, stampDict)
                 where
                   pilotsToSubmit is the pilots still to submit (maybe 0)
                   pilotsList is the list of pilots submitted
                   stampDict is a dict of timestamps of pilots submission
        :rtype: dict
        """
        self.log.info("Going to submit pilots", "(a maximum of %d pilots to %s queue)" % (pilotsToSubmit, queue))

        bundleProxy = self.queueDict[queue].get("BundleProxy", False)
        proxy = None
        if bundleProxy:
            proxy = ce.proxy

        jobExecDir = self.queueDict[queue]["ParametersDict"].get("JobExecDir", "")
        envVariables = self.queueDict[queue]["ParametersDict"].get("EnvironmentVariables", None)

        executable = self.getExecutable(queue, proxy=proxy, jobExecDir=jobExecDir, envVariables=envVariables)

        submitResult = ce.submitJob(executable, "", pilotsToSubmit)
        # In case the CE does not need the executable after the submission, we delete it
        # Else, we keep it, the CE will delete it after the end of the pilot execution
        if submitResult.get("ExecutableToKeep") != executable:
            os.unlink(executable)

        if not submitResult["OK"]:
            self.log.error("Failed submission to queue", "Queue %s:\n, %s" % (queue, submitResult["Message"]))

            if self.sendSubmissionAccounting:
                self.sendPilotSubmissionAccounting(
                    self.queueDict[queue]["Site"],
                    self.queueDict[queue]["CEName"],
                    self.queueDict[queue]["QueueName"],
                    pilotsToSubmit,
                    0,
                    "Failed",
                )

            self.failedQueues[queue] += 1
            return submitResult

        # Add pilots to the PilotAgentsDB: assign pilots to TaskQueue proportionally to the task queue priorities
        pilotList = submitResult["Value"]
        self.queueSlots[queue]["AvailableSlots"] -= len(pilotList)

        self.totalSubmittedPilots += len(pilotList)
        self.log.info(
            "Submitted %d pilots to %s@%s"
            % (len(pilotList), self.queueDict[queue]["QueueName"], self.queueDict[queue]["CEName"])
        )
        stampDict = submitResult.get("PilotStampDict", {})
        if self.sendSubmissionAccounting:
            self.sendPilotSubmissionAccounting(
                self.queueDict[queue]["Site"],
                self.queueDict[queue]["CEName"],
                self.queueDict[queue]["QueueName"],
                len(pilotList),
                len(pilotList),
                "Succeeded",
            )

        return S_OK((pilotList, stampDict))

    def _addPilotTQReference(self, queue, taskQueueDict, pilotList, stampDict):
        """Add to pilotAgentsDB the reference of for which TqID the pilots have been sent

        :param str queue: the queue name
        :param taskQueueDict: dict of task queues
        :type taskQueueDict: dict
        :param pilotList: list of pilots
        :type pilotList: list
        :param stampDict: dictionary of pilots timestamps
        :type stampDict: dict

        :return: None
        """

        tqPriorityList = []
        sumPriority = 0.0
        for tq in taskQueueDict:
            sumPriority += taskQueueDict[tq]["Priority"]
            tqPriorityList.append((tq, sumPriority))
        tqDict = {}
        for pilotID in pilotList:
            rndm = random.random() * sumPriority
            for tq, prio in tqPriorityList:
                if rndm < prio:
                    tqID = tq
                    break
            if tqID not in tqDict:
                tqDict[tqID] = []
            tqDict[tqID].append(pilotID)

        for tqID, pilotsList in tqDict.items():
            result = pilotAgentsDB.addPilotTQReference(
                pilotsList,
                tqID,
                self.pilotDN,
                self.pilotGroup,
                self.localhost,
                self.queueDict[queue]["CEType"],
                stampDict,
            )
            if not result["OK"]:
                self.log.error("Failed add pilots to the PilotAgentsDB", result["Message"])
                continue
            for pilot in pilotsList:
                result = pilotAgentsDB.setPilotStatus(
                    pilot,
                    PilotStatus.SUBMITTED,
                    self.queueDict[queue]["CEName"],
                    "Successfully submitted by the SiteDirector",
                    self.queueDict[queue]["Site"],
                    self.queueDict[queue]["QueueName"],
                )
                if not result["OK"]:
                    self.log.error("Failed to set pilot status", result["Message"])
                    continue

    def getQueueSlots(self, queue, manyWaitingPilotsFlag):
        """Get the number of available slots in the queue"""
        ce = self.queueDict[queue]["CE"]
        ceName = self.queueDict[queue]["CEName"]
        queueName = self.queueDict[queue]["QueueName"]
        queryCEFlag = self.queueDict[queue]["QueryCEFlag"].lower() in ["1", "yes", "true"]

        self.queueSlots.setdefault(queue, {})
        totalSlots = self.queueSlots[queue].get("AvailableSlots", 0)

        # See if there are waiting pilots for this queue. If not, allow submission
        if totalSlots and manyWaitingPilotsFlag:
            result = pilotAgentsDB.selectPilots(
                {"DestinationSite": ceName, "Queue": queueName, "Status": PilotStatus.PILOT_WAITING_STATES}
            )
            if result["OK"]:
                jobIDList = result["Value"]
                if not jobIDList:
                    return totalSlots
            return 0

        availableSlotsCount = self.queueSlots[queue].setdefault("AvailableSlotsCount", 0)
        waitingJobs = 1
        if totalSlots == 0:
            if availableSlotsCount % self.availableSlotsUpdateCycleFactor == 0:

                # Get the list of already existing pilots for this queue
                jobIDList = None
                result = pilotAgentsDB.selectPilots(
                    {"DestinationSite": ceName, "Queue": queueName, "Status": PilotStatus.PILOT_TRANSIENT_STATES}
                )

                if result["OK"]:
                    jobIDList = result["Value"]

                if queryCEFlag:
                    result = ce.available(jobIDList)
                    if not result["OK"]:
                        self.log.warn(
                            "Failed to check the availability of queue", "%s: \n%s" % (queue, result["Message"])
                        )
                        self.failedQueues[queue] += 1
                    else:
                        ceInfoDict = result["CEInfoDict"]
                        self.log.info(
                            "CE queue report",
                            "(%s_%s): Wait=%d, Run=%d, Submitted=%d, Max=%d"
                            % (
                                ceName,
                                queueName,
                                ceInfoDict["WaitingJobs"],
                                ceInfoDict["RunningJobs"],
                                ceInfoDict["SubmittedJobs"],
                                ceInfoDict["MaxTotalJobs"],
                            ),
                        )
                        totalSlots = result["Value"]
                        self.queueSlots[queue]["AvailableSlots"] = totalSlots
                        waitingJobs = ceInfoDict["WaitingJobs"]
                else:
                    maxWaitingJobs = int(self.queueDict[queue]["ParametersDict"].get("MaxWaitingJobs", 10))
                    maxTotalJobs = int(self.queueDict[queue]["ParametersDict"].get("MaxTotalJobs", 10))
                    waitingToRunningRatio = float(
                        self.queueDict[queue]["ParametersDict"].get("WaitingToRunningRatio", 0.0)
                    )
                    waitingJobs = 0
                    totalJobs = 0
                    if jobIDList:
                        result = pilotAgentsDB.getPilotInfo(jobIDList)
                        if not result["OK"]:
                            self.log.warn(
                                "Failed to check PilotAgentsDB", "for queue %s: \n%s" % (queue, result["Message"])
                            )
                            self.failedQueues[queue] += 1
                        else:
                            for _pilotRef, pilotDict in result["Value"].items():
                                if pilotDict["Status"] in PilotStatus.PILOT_TRANSIENT_STATES:
                                    totalJobs += 1
                                    if pilotDict["Status"] in PilotStatus.PILOT_WAITING_STATES:
                                        waitingJobs += 1
                            runningJobs = totalJobs - waitingJobs
                            self.log.info(
                                "PilotAgentsDB report",
                                "(%s_%s): Wait=%d, Run=%d, Max=%d"
                                % (ceName, queueName, waitingJobs, runningJobs, maxTotalJobs),
                            )
                            maxWaitingJobs = int(max(maxWaitingJobs, runningJobs * waitingToRunningRatio))

                    totalSlots = min((maxTotalJobs - totalJobs), (maxWaitingJobs - waitingJobs))
                    self.queueSlots[queue]["AvailableSlots"] = max(totalSlots, 0)

        self.queueSlots[queue]["AvailableSlotsCount"] += 1

        if manyWaitingPilotsFlag and waitingJobs:
            return 0
        return totalSlots

    #####################################################################################
    def getExecutable(self, queue, proxy=None, jobExecDir="", envVariables=None, **kwargs):
        """Prepare the full executable for queue

        :param str queue: queue name
        :param bundleProxy: flag that say if to bundle or not the proxy
        :type bundleProxy: bool
        :param str queue: pilot execution dir (normally an empty string)

        :returns: a string the options for the pilot
        :rtype: str
        """

        pilotOptions = self._getPilotOptions(queue, **kwargs)
        if not pilotOptions:
            self.log.warn("Pilots will be submitted without additional options")
            pilotOptions = []
        pilotOptions = " ".join(pilotOptions)
        self.log.verbose("pilotOptions: %s" % pilotOptions)

        # if a global workingDirectory is defined for the CEType (like HTCondor)
        # use it (otherwise the __cleanup done by HTCondor will be in the wrong folder !)
        # Note that this means that if you run multiple HTCondorCE
        # in your machine, the executable files will be in the same place
        # but it does not matter since they are very temporary

        ce = self.queueCECache[queue]["CE"]
        workingDirectory = getattr(ce, "workingDirectory", self.workingDirectory)

        executable = self._writePilotScript(
            workingDirectory=workingDirectory,
            pilotOptions=pilotOptions,
            proxy=proxy,
            pilotExecDir=jobExecDir,
            envVariables=envVariables,
        )
        return executable

    #####################################################################################

    def _getPilotOptions(self, queue, **kwargs):
        """Prepare pilot options

        :param str queue: queue name

        :returns: pilotOptions is a list of strings, each one is an option to the dirac-pilot script invocation
        :rtype: list
        """
        queueDict = self.queueDict[queue]["ParametersDict"]
        pilotOptions = []

        setup = gConfig.getValue("/DIRAC/Setup", "unknown")
        if setup == "unknown":
            self.log.error("Setup is not defined in the configuration")
            return [None, None]
        pilotOptions.append("-S %s" % setup)
        opsHelper = Operations(group=self.pilotGroup, setup=setup)

        # Installation defined?
        installationName = opsHelper.getValue("Pilot/Installation", "")
        if installationName:
            pilotOptions.append("-V %s" % installationName)

        # Project defined?
        projectName = opsHelper.getValue("Pilot/Project", "")
        if projectName:
            pilotOptions.append("-l %s" % projectName)
        else:
            self.log.info("DIRAC project will be installed by pilots")

        # Pilot Logging defined?
        pilotLogging = opsHelper.getValue("/Services/JobMonitoring/usePilotsLoggingFlag", False)
        if pilotLogging:
            pilotOptions.append("-z ")

        # python 3 pilots?
        if self.python3Pilots:
            pilotOptions.append("--pythonVersion=3")

        # Debug
        if self.pilotLogLevel.lower() == "debug":
            pilotOptions.append("-ddd")

        # DIRAC Extensions to be used in pilots
        pilotExtensionsList = opsHelper.getValue("Pilot/Extensions", [])
        extensionsList = []
        if pilotExtensionsList:
            if pilotExtensionsList[0] != "None":
                extensionsList = pilotExtensionsList
        else:
            extensionsList = [ext for ext in CSGlobals.getCSExtensions() if "Web" not in ext]
        if extensionsList:
            pilotOptions.append("-e %s" % ",".join(extensionsList))

        # CEName
        pilotOptions.append("-N %s" % self.queueDict[queue]["CEName"])
        # Queue
        pilotOptions.append("-Q %s" % self.queueDict[queue]["QueueName"])
        # SiteName
        pilotOptions.append("-n %s" % queueDict["Site"])

        if "SharedArea" in queueDict:
            pilotOptions.append("-o '/LocalSite/SharedArea=%s'" % queueDict["SharedArea"])

        if "ExtraPilotOptions" in queueDict:
            pilotOptions.append(queueDict["ExtraPilotOptions"])

        if self.group:
            pilotOptions.append("-G %s" % self.group)

        return pilotOptions

    ####################################################################################

    def _writePilotScript(self, workingDirectory, pilotOptions, proxy=None, pilotExecDir="", envVariables=None):
        """Bundle together and write out the pilot executable script, admix the proxy if given

        :param str workingDirectory: pilot wrapper working directory
        :param str pilotOptions: options with which to start the pilot
        :param str proxy: proxy file we are going to bundle
        :param str pilotExecDir: pilot executing directory

        :returns: file name of the pilot wrapper created
        :rtype: str
        """

        try:
            pilotFilesCompressedEncodedDict = getPilotFilesCompressedEncodedDict([], proxy)
        except Exception as be:
            self.log.exception("Exception during pilot modules files compression", lException=be)

        location = Operations().getValue("Pilot/pilotFileServer", "")
        localPilot = pilotWrapperScript(
            pilotFilesCompressedEncodedDict=pilotFilesCompressedEncodedDict,
            pilotOptions=pilotOptions,
            pilotExecDir=pilotExecDir,
            envVariables=envVariables,
            location=location,
        )

        return _writePilotWrapperFile(workingDirectory=workingDirectory, localPilot=localPilot)

    def updatePilotStatus(self):
        """Update status of pilots in transient and final states"""

        # Generate a proxy before feeding the threads to renew the ones of the CEs to perform actions
        result = gProxyManager.getPilotProxyFromDIRACGroup(self.pilotDN, self.pilotGroup, 23400)
        if not result["OK"]:
            return result
        proxy = result["Value"]

        # Getting the status of pilots in a queue implies the use of remote CEs and may lead to network latency
        # Threads aim at overcoming such issues and thus 1 thread per queue is created to
        # update the status of pilots in transient states
        with ThreadPoolExecutor(max_workers=len(self.queueDict)) as executor:
            for queue in self.queueDict:
                executor.submit(self._updatePilotStatusPerQueue, queue, proxy)

        # The pilot can be in Done state set by the job agent check if the output is retrieved
        for queue in self.queueDict:
            ce = self.queueDict[queue]["CE"]

            if not ce.isProxyValid(120)["OK"]:
                result = gProxyManager.getPilotProxyFromDIRACGroup(self.pilotDN, self.pilotGroup, 1000)
                if not result["OK"]:
                    return result
                proxy = result["Value"]
                ce.setProxy(proxy, 940)

            ceName = self.queueDict[queue]["CEName"]
            queueName = self.queueDict[queue]["QueueName"]
            ceType = self.queueDict[queue]["CEType"]
            siteName = self.queueDict[queue]["Site"]
            result = pilotAgentsDB.selectPilots(
                {
                    "DestinationSite": ceName,
                    "Queue": queueName,
                    "GridType": ceType,
                    "GridSite": siteName,
                    "OutputReady": "False",
                    "Status": PilotStatus.PILOT_FINAL_STATES,
                }
            )

            if not result["OK"]:
                self.log.error("Failed to select pilots", result["Message"])
                continue
            pilotRefs = result["Value"]
            if not pilotRefs:
                continue
            result = pilotAgentsDB.getPilotInfo(pilotRefs)
            if not result["OK"]:
                self.log.error("Failed to get pilots info from DB", result["Message"])
                continue
            pilotDict = result["Value"]
            if self.getOutput:
                for pRef in pilotRefs:
                    self._getPilotOutput(pRef, pilotDict, ce, ceName)

            # Check if the accounting is to be sent
            if self.sendAccounting:
                result = pilotAgentsDB.selectPilots(
                    {
                        "DestinationSite": ceName,
                        "Queue": queueName,
                        "GridType": ceType,
                        "GridSite": siteName,
                        "AccountingSent": "False",
                        "Status": PilotStatus.PILOT_FINAL_STATES,
                    }
                )

                if not result["OK"]:
                    self.log.error("Failed to select pilots", result["Message"])
                    continue
                pilotRefs = result["Value"]
                if not pilotRefs:
                    continue
                result = pilotAgentsDB.getPilotInfo(pilotRefs)
                if not result["OK"]:
                    self.log.error("Failed to get pilots info from DB", result["Message"])
                    continue
                pilotDict = result["Value"]
                result = self.sendPilotAccounting(pilotDict)
                if not result["OK"]:
                    self.log.error("Failed to send pilot agent accounting")

        return S_OK()

    def _updatePilotStatusPerQueue(self, queue, proxy):
        """Update status of pilots in transient state for a given queue

        :param queue: queue name
        :param proxy: proxy to check the pilot status and renewals
        """
        ce = self.queueDict[queue]["CE"]
        ceName = self.queueDict[queue]["CEName"]
        queueName = self.queueDict[queue]["QueueName"]
        ceType = self.queueDict[queue]["CEType"]
        siteName = self.queueDict[queue]["Site"]

        result = pilotAgentsDB.selectPilots(
            {
                "DestinationSite": ceName,
                "Queue": queueName,
                "GridType": ceType,
                "GridSite": siteName,
                "Status": PilotStatus.PILOT_TRANSIENT_STATES,
                "OwnerDN": self.pilotDN,
                "OwnerGroup": self.pilotGroup,
            }
        )
        if not result["OK"]:
            self.log.error('Failed to select pilots", ": %s' % result["Message"])
            return
        pilotRefs = result["Value"]
        if not pilotRefs:
            return

        result = pilotAgentsDB.getPilotInfo(pilotRefs)
        if not result["OK"]:
            self.log.error("Failed to get pilots info from DB", result["Message"])
            return
        pilotDict = result["Value"]

        stampedPilotRefs = []
        for pRef in pilotDict:
            if pilotDict[pRef]["PilotStamp"]:
                stampedPilotRefs.append(pRef + ":::" + pilotDict[pRef]["PilotStamp"])
            else:
                stampedPilotRefs = list(pilotRefs)
                break

        # This proxy is used for checking the pilot status and renewals
        # We really need at least a few hours otherwise the renewed
        # proxy may expire before we check again...
        result = ce.isProxyValid(3 * 3600)
        if not result["OK"]:
            ce.setProxy(proxy, 23300)

        result = ce.getJobStatus(stampedPilotRefs)
        if not result["OK"]:
            self.log.error("Failed to get pilots status from CE", "%s: %s" % (ceName, result["Message"]))
            return
        pilotCEDict = result["Value"]

        abortedPilots, getPilotOutput = self._updatePilotStatus(pilotRefs, pilotDict, pilotCEDict)
        for pRef in getPilotOutput:
            self._getPilotOutput(pRef, pilotDict, ce, ceName)

        # If something wrong in the queue, make a pause for the job submission
        if abortedPilots:
            self.failedQueues[queue] += 1

    def _updatePilotStatus(self, pilotRefs, pilotDict, pilotCEDict):
        """Really updates the pilots status

        :return: number of aborted pilots, flag for getting the pilot output
        """

        abortedPilots = 0
        getPilotOutput = []

        for pRef in pilotRefs:
            newStatus = ""
            oldStatus = pilotDict[pRef]["Status"]
            lastUpdateTime = pilotDict[pRef]["LastUpdateTime"]
            sinceLastUpdate = dateTime() - lastUpdateTime

            ceStatus = pilotCEDict.get(pRef, oldStatus)

            if oldStatus == ceStatus and ceStatus != PilotStatus.UNKNOWN:
                # Normal status did not change, continue
                continue
            elif ceStatus == oldStatus == PilotStatus.UNKNOWN:
                if sinceLastUpdate < 3600 * second:
                    # Allow 1 hour of Unknown status assuming temporary problems on the CE
                    continue
                else:
                    newStatus = PilotStatus.ABORTED
            elif ceStatus == PilotStatus.UNKNOWN and oldStatus not in PilotStatus.PILOT_FINAL_STATES:
                # Possible problems on the CE, let's keep the Unknown status for a while
                newStatus = PilotStatus.UNKNOWN
            elif ceStatus != PilotStatus.UNKNOWN:
                # Update the pilot status to the new value
                newStatus = ceStatus

            if newStatus:
                self.log.info("Updating status", "to %s for pilot %s" % (newStatus, pRef))
                result = pilotAgentsDB.setPilotStatus(pRef, newStatus, "", "Updated by SiteDirector")
                if not result["OK"]:
                    self.log.error(result["Message"])
                if newStatus == "Aborted":
                    abortedPilots += 1
            # Set the flag to retrieve the pilot output now or not
            if newStatus in PilotStatus.PILOT_FINAL_STATES:
                if pilotDict[pRef]["OutputReady"].lower() == "false" and self.getOutput:
                    getPilotOutput.append(pRef)

        return abortedPilots, getPilotOutput

    def _getPilotOutput(self, pRef, pilotDict, ce, ceName):
        """Retrieves the pilot output for a pilot and stores it in the pilotAgentsDB"""
        self.log.info("Retrieving output for pilot %s" % pRef)
        output = None
        error = None

        pilotStamp = pilotDict[pRef]["PilotStamp"]
        pRefStamp = pRef
        if pilotStamp:
            pRefStamp = pRef + ":::" + pilotStamp

        result = ce.getJobOutput(pRefStamp)
        if not result["OK"]:
            self.failedPilotOutput[pRefStamp] += 1
            self.log.error("Failed to get pilot output", "%s: %s" % (ceName, result["Message"]))
            self.log.verbose(
                "Retries left: %d" % max(0, self.maxRetryGetPilotOutput - self.failedPilotOutput[pRefStamp])
            )

            if (self.maxRetryGetPilotOutput - self.failedPilotOutput[pRefStamp]) <= 0:
                output = "Output is no longer available"
                error = "Error is no longer available"
                self.failedPilotOutput.pop(pRefStamp)
            else:
                return
        else:
            output, error = result["Value"]

        if output:
            result = pilotAgentsDB.storePilotOutput(pRef, output, error)
            if not result["OK"]:
                self.log.error("Failed to store pilot output", result["Message"])
        else:
            self.log.warn("Empty pilot output not stored to PilotDB")

    def sendPilotAccounting(self, pilotDict):
        """Send pilot accounting record"""
        for pRef in pilotDict:
            self.log.verbose("Preparing accounting record", "for pilot %s" % pRef)
            pA = PilotAccounting()
            pA.setEndTime(pilotDict[pRef]["LastUpdateTime"])
            pA.setStartTime(pilotDict[pRef]["SubmissionTime"])
            retVal = Registry.getUsernameForDN(pilotDict[pRef]["OwnerDN"])
            if not retVal["OK"]:
                userName = "unknown"
                self.log.error("Can't determine username for dn", pilotDict[pRef]["OwnerDN"])
            else:
                userName = retVal["Value"]
            pA.setValueByKey("User", userName)
            pA.setValueByKey("UserGroup", pilotDict[pRef]["OwnerGroup"])
            result = getCESiteMapping(pilotDict[pRef]["DestinationSite"])
            if result["OK"] and result["Value"]:
                pA.setValueByKey("Site", result["Value"][pilotDict[pRef]["DestinationSite"]].strip())
            else:
                pA.setValueByKey("Site", "Unknown")
            pA.setValueByKey("GridCE", pilotDict[pRef]["DestinationSite"])
            pA.setValueByKey("GridMiddleware", pilotDict[pRef]["GridType"])
            pA.setValueByKey("GridResourceBroker", pilotDict[pRef]["Broker"])
            pA.setValueByKey("GridStatus", pilotDict[pRef]["Status"])
            if "Jobs" not in pilotDict[pRef]:
                pA.setValueByKey("Jobs", 0)
            else:
                pA.setValueByKey("Jobs", len(pilotDict[pRef]["Jobs"]))
            self.log.verbose("Adding accounting record", "for pilot %s" % pilotDict[pRef]["PilotID"])
            retVal = gDataStoreClient.addRegister(pA)
            if not retVal["OK"]:
                self.log.error("Failed to send accounting info for pilot ", pRef)
            else:
                # Set up AccountingSent flag
                result = pilotAgentsDB.setAccountingFlag(pRef)
                if not result["OK"]:
                    self.log.error("Failed to set accounting flag for pilot ", pRef)

        self.log.info("Committing accounting records", "for %d pilots" % len(pilotDict))
        result = gDataStoreClient.commit()
        if result["OK"]:
            for pRef in pilotDict:
                self.log.verbose("Setting AccountingSent flag", "for pilot %s" % pRef)
                result = pilotAgentsDB.setAccountingFlag(pRef)
                if not result["OK"]:
                    self.log.error("Failed to set accounting flag for pilot ", pRef)
        else:
            return result

        return S_OK()

    def sendPilotSubmissionAccounting(self, siteName, ceName, queueName, numTotal, numSucceeded, status):
        """Send pilot submission accounting record

        :param str siteName:     Site name
        :param str ceName:       CE name
        :param str queueName:    queue Name
        :param int numTotal:     Total number of submission
        :param int numSucceeded: Total number of submission succeeded
        :param str status:       'Succeeded' or 'Failed'

        :returns: S_OK / S_ERROR
        """

        pA = PilotSubmissionAccounting()
        pA.setStartTime(dateTime())
        pA.setEndTime(dateTime())
        pA.setValueByKey("HostName", DIRAC.siteName())
        if hasattr(self, "_AgentModule__moduleProperties"):
            pA.setValueByKey("SiteDirector", self.am_getModuleParam("agentName"))
        else:  # In case it is not executed as agent
            pA.setValueByKey("SiteDirector", "Client")

        pA.setValueByKey("Site", siteName)
        pA.setValueByKey("CE", ceName)
        pA.setValueByKey("Queue", ceName + ":" + queueName)
        pA.setValueByKey("Status", status)
        pA.setValueByKey("NumTotal", numTotal)
        pA.setValueByKey("NumSucceeded", numSucceeded)
        result = gDataStoreClient.addRegister(pA)

        if not result["OK"]:
            self.log.warn("Error in add Register:" + result["Message"])
            return result

        self.log.verbose("Begin commit")
        result = gDataStoreClient.delayedCommit()
        if not result["OK"]:
            self.log.error("Error in Commit:" + result["Message"])
            return result



""" A set of utilities for getting configuration information for the WMS components
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.ConfigurationSystem.Client.Helpers import Registry, Operations


def findGenericPilotCredentials(vo=False, group=False, pilotDN="", pilotGroup=""):
    """Looks into the Operations/<>/Pilot section of CS to find the pilot credentials.
    Then check if the user has a registered proxy in ProxyManager.

    if pilotDN or pilotGroup are specified, use them

    :param str vo: VO name
    :param str group: group name
    :param str pilotDN: pilot DN
    :param str pilotGroup: pilot group

    :return: S_OK(tuple)/S_ERROR()
    """
    if not group and not vo:
        return S_ERROR("Need a group or a VO to determine the Generic pilot credentials")
    if not vo:
        vo = Registry.getVOForGroup(group)
        if not vo:
            return S_ERROR("Group %s does not have a VO associated" % group)
    opsHelper = Operations.Operations(vo=vo)
    if not pilotGroup:
        pilotGroup = opsHelper.getValue("Pilot/GenericPilotGroup", "")
    if not pilotDN:
        pilotDN = opsHelper.getValue("Pilot/GenericPilotDN", "")
    if not pilotDN:
        pilotUser = opsHelper.getValue("Pilot/GenericPilotUser", "")
        if pilotUser:
            result = Registry.getDNForUsername(pilotUser)
            if result["OK"]:
                pilotDN = result["Value"][0]
    if pilotDN and pilotGroup:
        gLogger.verbose("Pilot credentials: %s@%s" % (pilotDN, pilotGroup))
        result = gProxyManager.userHasProxy(pilotDN, pilotGroup, 86400)
        if not result["OK"]:
            return S_ERROR("%s@%s has no proxy in ProxyManager")
        return S_OK((pilotDN, pilotGroup))

    if pilotDN:
        return S_ERROR("DN %s does not have group %s" % (pilotDN, pilotGroup))
    return S_ERROR("No generic proxy in the Proxy Manager with groups %s" % pilotGroup)



""" Create and send a combined request for any pending operations at
    the end of a job:

      * fileReport (for the transformation)
      * jobReport (for jobs)
      * accounting
      * request (for failover)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.Workflow.Modules.ModuleBase import ModuleBase, GracefulTermination


class FailoverRequest(ModuleBase):

    #############################################################################

    def __init__(self):
        """Module initialization."""

        self.log = gLogger.getSubLogger("FailoverRequest")
        super(FailoverRequest, self).__init__(self.log)

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module input parameters are resolved here."""
        super(FailoverRequest, self)._resolveInputVariables()
        super(FailoverRequest, self)._resolveInputStep()

    def _initialize(self):
        """checks if is to do, then prepare few things"""
        if not self._enableModule():
            raise GracefulTermination("Skipping FailoverRequest module")

        self.request.RequestName = "job_%d_request.xml" % self.jobID
        self.request.JobID = self.jobID
        self.request.SourceComponent = "Job_%d" % self.jobID

    def _execute(self):

        # report on the status of the input data, by default they are 'Processed', unless the job failed
        # failures happening before are not touched
        filesInFileReport = self.fileReport.getFiles()
        if not self._checkWFAndStepStatus(noPrint=True):
            for lfn in self.inputDataList:
                if lfn not in filesInFileReport:
                    self.log.info("Forcing status to 'Unused' due to workflow failure for: %s" % (lfn))
                    # Set the force flag in case the file was in a terminal status
                    self.fileReport.force = True
                    self.fileReport.setFileStatus(int(self.production_id), lfn, TransformationFilesStatus.UNUSED)
        else:
            filesInFileReport = self.fileReport.getFiles()
            for lfn in self.inputDataList:
                if lfn not in filesInFileReport:
                    self.log.verbose("No status populated for input data %s, setting to 'Processed'" % lfn)
                    self.fileReport.setFileStatus(int(self.production_id), lfn, TransformationFilesStatus.PROCESSED)

        result = self.fileReport.commit()
        if not result["OK"]:
            self.log.error("Failed to report file status to TransformationDB")
            self.log.error("Trying again before populating request with file report information")
            result = self.fileReport.generateForwardDISET()
            if not result["OK"]:
                self.log.warn("Could not generate Operation for file report with result:\n%s" % (result["Value"]))
            else:
                if result["Value"] is None:
                    self.log.info("Files correctly reported to TransformationDB")
                else:
                    result = self.request.addOperation(result["Value"])
        else:
            self.log.info("Status of files have been properly updated in the TransformationDB")

        # Must ensure that the local job report instance is used to report the final status
        # in case of failure and a subsequent failover operation
        if self.workflowStatus["OK"] and self.stepStatus["OK"]:
            self.setApplicationStatus("Job Finished Successfully")

            self.generateFailoverFile()

    def _finalize(self):
        """Finalize and report correct status for the workflow based on the workflow
        or step status.
        """
        if not self._checkWFAndStepStatus(True):
            raise RuntimeError("Workflow failed, FailoverRequest module completed")

        super(FailoverRequest, self)._finalize("Workflow successful, end of FailoverRequest module execution.")
